"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Aspect calculations are compared against those found by astro.com.
    Where possible, astro.com's aspects table is the preferred source
    of data, but the chart visual is also used to confirm data not in
    the table.

"""

import math

from datetime import datetime

from pytest import fixture

from immanuel.const import calc, chart
from immanuel.reports import aspect
from immanuel.setup import settings
from immanuel.tools import convert, date, ephemeris


@fixture
def objects():
    lat, lon = [convert.string_to_dec(v) for v in ('32n43', '117w09')]
    jd = date.to_jd('2000-01-01 10:00', lat, lon)
    return ephemeris.objects(settings.objects, jd, lat, lon, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)

@fixture
def partner_objects():
    lat, lon = [convert.string_to_dec(v) for v in ('38n35', '121w30')]
    jd = date.to_jd('2001-02-16 06:00', lat, lon)
    return ephemeris.objects(settings.objects, jd, lat, lon, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)


def test_between(objects):
    a = aspect.between(objects[chart.SUN], objects[chart.MOON])
    assert a['active'] == chart.MOON
    assert a['passive'] == chart.SUN
    assert a['aspect'] == calc.SEXTILE
    assert a['difference'] <= a['orb']
    assert math.ceil(a['difference']) == -5                 # astro.com rounds up
    assert a['distance'] - a['difference'] == calc.SEXTILE
    assert a['movement'] == calc.SEPARATIVE
    assert a['condition'] == calc.ASSOCIATE                 # Not on astro.com report, can be ascertained visually


def test_for_object(objects):
    settings.aspect_rules[chart.ASC] = settings.default_aspect_rule     # astro.com chart visual does not include aspects to Asc but its aspects table does
    a = aspect.for_object(objects[chart.SUN], objects)
    assert sorted(tuple(a.keys())) == sorted((chart.ASC, chart.PARS_FORTUNA, chart.MOON, chart.MERCURY, chart.SATURN))
    assert a[chart.ASC]['aspect'] == calc.SEXTILE
    assert a[chart.PARS_FORTUNA]['aspect'] == calc.CONJUNCTION
    assert a[chart.MOON]['aspect'] == calc.SEXTILE
    assert a[chart.MERCURY]['aspect'] == calc.CONJUNCTION
    assert a[chart.SATURN]['aspect'] == calc.TRINE


def test_all(objects):
    settings.aspect_rules[chart.ASC] = settings.default_aspect_rule     # astro.com chart visual does not include aspects to Asc but its aspects table does
    settings.aspect_rules[chart.DESC] = {                               # Because we're looking at Asc aspects, don't allow Desc any
        'initiate': (),
        'receive': (),
    }
    all = aspect.all(objects)
    for index, aspect_list in all.items():
        for i, a in aspect_list.items():
            assert i in all
            assert index in all[i]
            assert a['aspect'] == all[i][index]['aspect']


def test_by_type(objects):
    settings.aspect_rules[chart.ASC] = settings.default_aspect_rule                                     # astro.com chart visual does not include aspects to Asc but its aspects table does
    settings.aspects = (calc.CONJUNCTION, calc.OPPOSITION, calc.SQUARE, calc.TRINE, calc.SEXTILE)       # Copy astro.com's "only major aspects"
    by_type = aspect.by_type(objects)
    assert sorted(tuple(by_type.keys())) == sorted((calc.CONJUNCTION, calc.OPPOSITION, calc.SQUARE, calc.TRINE, calc.SEXTILE))


def test_synastry(objects, partner_objects):
    synastry = aspect.synastry(objects, partner_objects)
    # Due to the sheer number of aspects between 2 charts, we spot-check the sun against astro.com
    sun = synastry[chart.SUN]
    assert chart.TRUE_NORTH_NODE in sun
    assert sun[chart.TRUE_NORTH_NODE]['aspect'] == calc.OPPOSITION
    assert chart.TRUE_SOUTH_NODE in sun
    assert sun[chart.TRUE_SOUTH_NODE]['aspect'] == calc.CONJUNCTION
    assert chart.VENUS in sun
    assert sun[chart.VENUS]['aspect'] == calc.SQUARE



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module provides conversions between D:M:S and decimal.

    These functions perform simple conversions between the base-12 data used
    by astrology (ie. angles, coordinates, and time) in string/list format,
    and decimal numbers in float format.

"""

import math
import re

import swisseph as swe


FORMAT_TIME = 0
FORMAT_TIME_OFFSET = 1
FORMAT_DMS = 2
FORMAT_LAT = 3
FORMAT_LON = 4

ROUND_DEGREE = (1, swe.SPLIT_DEG_ROUND_DEG)
ROUND_MINUTE = (2, swe.SPLIT_DEG_ROUND_MIN)
ROUND_SECOND = (3, swe.SPLIT_DEG_ROUND_SEC)


def dms_to_dec(dms: list | tuple) -> float:
    """ Returns the decimal conversion of a D:M:S list. """
    dec = sum([float(abs(v)) / 60**k for k, v in enumerate(dms[1:])])
    return dec if dms[0] == '+' else -dec


def dec_to_dms(dec: float, round_to: tuple = ROUND_SECOND, pad_rounded: bool = False) -> tuple:
    """ Returns the rounded D:M:S conversion of a decimal float. """
    dms = ('-' if dec < 0 else '+', *swe.split_deg(dec, round_to[1])[:round_to[0]])
    return dms + (0,) * (3-round_to[0]) if pad_rounded else dms


def dms_to_string(dms: list | tuple, format: int = FORMAT_DMS, round_to: tuple = ROUND_SECOND, pad_rounded: bool = None) -> str:
    """ Returns a D:M:S list as either a D:M:S, D°M'S" or
    lat/lon coordinate string. """
    pad_rounded = True if format in (FORMAT_LAT, FORMAT_LON) or (pad_rounded is None and format != FORMAT_DMS) else pad_rounded
    dms = dec_to_dms(dms_to_dec(dms), round_to, pad_rounded)

    if format == FORMAT_DMS:
        return _dms_to_string_format_dms(dms)
    if format == FORMAT_TIME:
        return _dms_to_string_format_time(dms)
    if format == FORMAT_TIME_OFFSET:
        return _dms_to_string_format_time_offset(dms)
    if format == FORMAT_LAT:
        return _dms_to_string_format_lat(dms)
    if format == FORMAT_LON:
        return _dms_to_string_format_lon(dms)

    return ''


def string_to_dms(string: str, round_to: tuple = ROUND_SECOND, pad_rounded: bool = False) -> tuple:
    """ Takes any string supported by string_to_dec() and returns a
    DMS tuple. """
    return dec_to_dms(string_to_dec(string), round_to, pad_rounded)


def dec_to_string(dec: float, format: int = FORMAT_DMS, round_to: tuple = ROUND_SECOND, pad_rounded: bool = None) -> str:
    """ Returns a decimal float as either a D:M:S or a D°M'S" string. """
    return dms_to_string(dec_to_dms(dec, round_to), format, round_to, pad_rounded)


def string_to_dec(string: str) -> float:
    """ Takes any string format output by dms_to_string() and returns
    a decimal float. """
    digits = re.findall(r'[0-9\.-]+', string)
    char = string[len(digits[0])].upper()
    floats = [float(v) for v in digits]
    return dms_to_dec(['-' if char in 'SW' or floats[0] < 0 else '+', *floats])


def to_dec(value: float | list | tuple | str) -> float:
    """ If the input type is unknown, this will guess and convert. """
    if isinstance(value, float):
        return value
    if isinstance(value, (list, tuple)):
        return dms_to_dec(value)
    if isinstance(value, str):
        if _is_numeric(value):
            return float(value)
        else:
            return string_to_dec(value)
    return None


def to_dms(value: float | list | tuple | str, round_to: tuple = ROUND_SECOND, pad_rounded = False) -> tuple:
    """ If the input type is unknown, this will guess and convert. """
    if isinstance(value, float):
        return dec_to_dms(value, round_to, pad_rounded)
    if isinstance(value, (list, tuple)):
        return tuple(value)
    if isinstance(value, str):
        if _is_numeric(value):
            return dec_to_dms(float(value), round_to, pad_rounded)
        else:
            return string_to_dms(value, round_to, pad_rounded)
    return None


def to_string(value: float | list | tuple | str, format: int = FORMAT_DMS, round_to: tuple = ROUND_SECOND, pad_rounded: bool = None) -> str:
    """ If the input type is unknown, this will guess and convert. """
    if isinstance(value, float):
        return dec_to_string(value, format, round_to, pad_rounded)
    if isinstance(value, (list, tuple)):
        return dms_to_string(value, format, round_to, pad_rounded)
    if isinstance(value, str):
        if _is_numeric(value):
            return dec_to_string(float(value), format, round_to, pad_rounded)
        else:
            return dec_to_string(string_to_dec(value), format, round_to, pad_rounded)
    return None


def _dms_to_string_format_dms(dms: list | tuple) -> str:
    """ Returns DMS in degree/minute/second format. """
    symbols = (u'\N{DEGREE SIGN}', "'", '"')
    string = ''.join([f'{v:02d}' + symbols[k] for k, v in enumerate(dms[1:])])
    return '-' + string if dms[0] == '-' else string


def _dms_to_string_format_time(dms: list | tuple) -> str:
    """ Returns DMS in hour:minute:second format. """
    string = ':'.join([f'{v:02d}' for v in dms[1:]])
    return '-' + string if dms[0] == '-' else string


def _dms_to_string_format_time_offset(dms: list | tuple) -> str:
    """ Returns DMS in signed hour:minute:second format. """
    return dms[0] + ':'.join([f'{v:02d}' for v in dms[1:]])


def _dms_to_string_format_lat(dms: list | tuple) -> str:
    """ Returns DMS in degree/direction/minute format. """
    dir = 'S' if dms[0] == '-' else 'N'
    return _dms_to_string_format_lat_lon(dms, dir)


def _dms_to_string_format_lon(dms: list | tuple) -> str:
    """ Returns DMS in degree/direction/minute format. """
    dir = 'W' if dms[0] == '-' else 'E'
    return _dms_to_string_format_lat_lon(dms, dir)


def _dms_to_string_format_lat_lon(dms: list | tuple, dir: str) -> str:
    """ Returns DMS in degree/direction/minute format. """
    minutes = dms[2] + math.ceil(((dms[3]/60)*100))/100
    return f'{dms[1]}{dir}{minutes}'


def _is_numeric(value: str) -> bool:
    """ Determine whether a string is numeric. """
    return re.match(r'^-?\d+(?:\.\d+)?$', value)



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Test the basic midpoint calculation methods against
    astro.com's midpoint composite chart.

"""

from datetime import datetime

import swisseph as swe
from pytest import fixture

from immanuel.const import calc, chart
from immanuel.tools import convert, date, ephemeris, midpoint, position


@fixture
def coords():
    # San Diego coords as used by astro.com
    return [convert.string_to_dec(v) for v in ('32n43', '117w09')]

@fixture
def jd1(coords):
    return date.to_jd('2000-01-01 10:00', *coords)

@fixture
def jd2(coords):
    return date.to_jd('2025-06-21 00:00', *coords)

@fixture
def obliquity(jd1, jd2):
    return midpoint.obliquity(jd1, jd2)

@fixture
def astro():
    return {
        chart.ASC: {
            'sign': chart.PISCES,
            'lon': '08°31\'36"',
            'dec': '-08°22\'21"',
        },
        chart.SUN: {
            'sign': chart.ARIES,
            'lon': '05°23\'50"',
            'lat': '00°00\'00"',
            'speed': '00°59\'13"',
            'dec': '02°08\'39"',
        },
        chart.MOON: {
            'sign': chart.AQUARIUS,
            'lon': '09°42\'33"',
            'lat': '00°00\'00"',
            'speed': '13°15\'20"',
            'dec': '-17°49\'05"',
        },
        chart.MERCURY: {
            'sign': chart.LIBRA,
            'lon': '12°12\'13"',
            'lat': '00°00\'00"',
            'speed': '01°33\'06"',
            'dec': '-04°49\'23"',
        },
        chart.VENUS: {
            'sign': chart.AQUARIUS,
            'lon': '23°42\'11"',
            'lat': '00°00\'00"',
            'speed': '01°08\'08"',
            'dec': '-13°37\'07"',
        },
        chart.MARS: {
            'sign': chart.SAGITTARIUS,
            'lon': '00°11\'05"',
            'lat': '00°00\'00"',
            'speed': '00°40\'14"',
            'dec': '-20°11\'18"',
        },
        chart.JUPITER: {
            'sign': chart.TAURUS,
            'lon': '28°55\'54"',
            'lat': '00°00\'00"',
            'speed': '00°08\'06"',
            'dec': '19°55\'10"',
        },
        chart.SATURN: {
            'sign': chart.ARIES,
            'lon': '20°57\'45"',
            'lat': '00°00\'00"',
            'speed': '00°00\'30"',
            'dec': '08°10\'52"',
        },
        chart.URANUS: {
            'sign': chart.ARIES,
            'lon': '07°01\'20"',
            'lat': '00°00\'00"',
            'speed': '00°03\'04"',
            'dec': '02°47\'14"',
        },
        chart.NEPTUNE: {
            'sign': chart.PISCES,
            'lon': '02°39\'48"',
            'lat': '00°00\'00"',
            # This is the only figure disagreeing with astro.com (~1 arcsec) and nobody knows why
            # 'speed': '00°01\'18"',
            'dec': '-10°31\'28"',
        },
        chart.PLUTO: {
            'sign': chart.CAPRICORN,
            'lon': '07°23\'57"',
            'lat': '00°00\'00"',
            'speed': '00°00\'29"',
            'dec': '-23°13\'53"',
        },
        chart.TRUE_NORTH_NODE: {
            'sign': chart.TAURUS,
            'lon': '28°04\'05"',
            'lat': '00°00\'00"',
            'speed': '-00°03\'45"',
            'dec': '19°43\'43"',
        },
        chart.PARS_FORTUNA: {
            'sign': chart.PISCES,
            'lon': '09°54\'57"',
            'lat': '00°00\'00"',
            'speed': '00°00\'00"',
            'dec': '-07°51\'02"',
        },
    }


def test_all(coords, jd1, jd2, obliquity, astro):
    objects1 = ephemeris.objects(astro.keys(), jd1, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
    objects2 = ephemeris.objects(astro.keys(), jd2, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
    composites = midpoint.all(objects1, objects2, obliquity)

    for index, composite in composites.items():
        sign = position.sign(composite)
        sign_lon = position.sign_longitude(composite)
        assert sign == astro[index]['sign']
        assert convert.dec_to_string(sign_lon) == astro[index]['lon']

        for key in ('lat', 'speed', 'dec'):
            if key in astro[index] and key in composite:
                assert convert.dec_to_string(composite[key]) == astro[index][key]


def test_composite(coords, jd1, jd2, obliquity, astro):
    for index in astro.keys():
        object1 = ephemeris.get(index, jd1, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
        object2 = ephemeris.get(index, jd2, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
        composite = midpoint.composite(object1, object2, obliquity)
        sign = position.sign(composite)
        sign_lon = position.sign_longitude(composite)
        assert sign == astro[index]['sign']
        assert convert.dec_to_string(sign_lon) == astro[index]['lon']

        for key in ('lat', 'speed', 'dec'):
            if key in astro[index] and key in composite:
                assert convert.dec_to_string(composite[key]) == astro[index][key]


def test_obliquity(jd1, jd2):
    obliquity = midpoint.obliquity(jd1, jd2, False)
    mean_obliquity = midpoint.obliquity(jd1, jd2, True)

    ecl_nut1 = swe.calc_ut(jd1, swe.ECL_NUT)[0]
    ecl_nut2 = swe.calc_ut(jd2, swe.ECL_NUT)[0]

    assert obliquity == (ecl_nut1[0] + ecl_nut2[0]) / 2
    assert mean_obliquity == (ecl_nut1[1] + ecl_nut2[1]) / 2



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module provides easy access to fairly standardised pyswisseph data.

    Relevant data on the main angles, houses, points and planets are
    available using the module's functions, many of which are cached.

    Many of the functions here, including angle, house and vertex functions,
    have an "armc_"-prefixed alternative if they are required to calculate
    from an ARMC.

"""

from functools import cache

import swisseph as swe

from immanuel.const import chart, names
from immanuel.tools import calculate, find


ALL = -1

_SWE = {
    chart.ALCABITUS: b'B',
    chart.AZIMUTHAL: b'H',
    chart.CAMPANUS: b'C',
    chart.EQUAL: b'A',
    chart.KOCH: b'K',
    chart.MERIDIAN: b'X',
    chart.MORINUS: b'M',
    chart.PLACIDUS: b'P',
    chart.POLICH_PAGE: b'T',
    chart.PORPHYRIUS: b'O',
    chart.REGIOMONTANUS: b'R',
    chart.VEHLOW_EQUAL: b'V',
    chart.WHOLE_SIGN: b'W',

    chart.ASC: swe.ASC,
    chart.DESC: swe.ASC,
    chart.MC: swe.MC,
    chart.IC: swe.MC,
    chart.ARMC: swe.ARMC,

    chart.SUN: swe.SUN,
    chart.MOON: swe.MOON,
    chart.MERCURY: swe.MERCURY,
    chart.VENUS: swe.VENUS,
    chart.MARS: swe.MARS,
    chart.JUPITER: swe.JUPITER,
    chart.SATURN: swe.SATURN,
    chart.URANUS: swe.URANUS,
    chart.NEPTUNE: swe.NEPTUNE,
    chart.PLUTO: swe.PLUTO,
    chart.CHIRON: swe.CHIRON,
    chart.PHOLUS: swe.PHOLUS,
    chart.CERES: swe.CERES,
    chart.PALLAS: swe.PALLAS,
    chart.JUNO: swe.JUNO,
    chart.VESTA: swe.VESTA,

    chart.NORTH_NODE: swe.MEAN_NODE,
    chart.SOUTH_NODE: swe.MEAN_NODE,
    chart.TRUE_NORTH_NODE: swe.TRUE_NODE,
    chart.TRUE_SOUTH_NODE: swe.TRUE_NODE,
    chart.VERTEX: swe.VERTEX,
    chart.LILITH: swe.MEAN_APOG,
    chart.TRUE_LILITH: swe.OSCU_APOG,
    chart.SYZYGY: chart.SYZYGY,
    chart.PARS_FORTUNA: chart.PARS_FORTUNA,
}


def objects(object_list: tuple, jd: float, lat: float = None, lon: float = None, house_system: int = None, pars_fortuna_formula: int = None) -> dict:
    """ Helper function returns a dict of all passed chart objects. """
    return _objects(
        object_list=object_list,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        pars_fortuna_formula=pars_fortuna_formula,
        armc=None,
        armc_obliquity=None
    )


def armc_objects(object_list: tuple, jd: float, armc: float, lat: float = None, lon: float = None, obliquity: float = None, house_system: int = None, pars_fortuna_formula: int = None) -> dict:
    """ Helper function returns a dict of all passed chart objects
    with points & angles calculated from the passed ARMC. """
    return _objects(
        object_list=object_list,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        pars_fortuna_formula=pars_fortuna_formula,
        armc=armc,
        armc_obliquity=obliquity
    )


def get(index: int | str, jd: float, lat: float = None, lon: float = None, house_system: int = None, pars_fortuna_formula: int = None) -> dict:
    """ Helper function to retrieve an angle, house, planet, point,
    asteroid, or fixed star. """
    return _get(
        index=index,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        pars_fortuna_formula=pars_fortuna_formula,
        armc=None,
        armc_obliquity=None
    )


def armc_get(index: int | str, jd: float, armc: float, lat: float = None, lon: float = None, obliquity: float = None, house_system: int = None, pars_fortuna_formula: int = None) -> dict:
    """ Helper function to retrieve an angle, house, planet, point,
    asteroid, or fixed star with houses & angles calculated from the
    passed ARMC. """
    return _get(
        index=index,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        pars_fortuna_formula=pars_fortuna_formula,
        armc=armc,
        armc_obliquity=obliquity
    )


def angles(jd: float, lat: float, lon: float, house_system: int) -> dict:
    """ Returns all four main chart angles & ARMC. """
    return _angle(
        index=ALL,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        armc=None,
        armc_obliquity=None
    )


def armc_angles(armc: float, lat: float, obliquity: float, house_system: int) -> dict:
    """ Returns all four main chart angles calculated from the
    passed ARMC. """
    return _angle(
        index=ALL,
        jd=None,
        lat=lat,
        lon=None,
        house_system=house_system,
        armc=armc,
        armc_obliquity=obliquity
    )


def angle(index: int, jd: float, lat: float, lon: float, house_system: int) -> dict:
    """ Returns one of the four main chart angles & its speed. Also stores
    the ARMC for further calculations. Returns all if index == ALL. """
    return _angle(
        index=index,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        armc=None,
        armc_obliquity=None
    )


def armc_angle(index: int, armc: float, lat: float, obliquity: float, house_system: int) -> dict:
    """ Returns one of the four main chart angles & its speed, calculated from
    the passed ARMC. Returns all if index == ALL. """
    return _angle(
        index=index,
        jd=None,
        lat=lat,
        lon=None,
        house_system=house_system,
        armc=armc,
        armc_obliquity=obliquity
    )


def houses(jd: float, lat: float, lon: float, house_system: int) -> dict:
    """ Returns all houses. """
    return _house(
        index=ALL,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        armc=None,
        armc_obliquity=None
    )


def armc_houses(armc: float, lat: float, obliquity: float, house_system: int) -> dict:
    """ Returns all houses calculated from the passed ARMC. """
    return _house(
        index=ALL,
        jd=None,
        lat=lat,
        lon=None,
        house_system=house_system,
        armc=armc,
        armc_obliquity=obliquity
    )


def house(index: int, jd: float, lat: float, lon: float, house_system: int) -> dict:
    """ Returns a house cusp & its speed, or all houses if index == ALL. """
    return _house(
        index=index,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        armc=None,
        armc_obliquity=None
    )


def armc_house(index: int, armc: float, lat: float, obliquity: float, house_system: int) -> dict:
    """ Returns a house cusp & its speed, or all houses if index == ALL,
    calculated from passed ARMC. """
    return _house(
        index=index,
        jd=None,
        lat=lat,
        lon=None,
        house_system=house_system,
        armc=armc,
        armc_obliquity=obliquity
    )


def point(index: int, jd: float, lat: float = None, lon: float = None, house_system: int = None, pars_fortuna_formula: int = None) -> dict:
    """ Returns a calculated point by Julian date, and additionally by lat / lon
    coordinates. """
    return _point(
        index=index,
        jd=jd,
        lat=lat,
        lon=lon,
        house_system=house_system,
        pars_fortuna_formula=pars_fortuna_formula,
        armc=None,
        armc_obliquity=None
    )


def armc_point(index: int, jd: float, armc: float, lat: float, obliquity: float, house_system: int = None, pars_fortuna_formula: int = None) -> dict:
    """ Returns a calculated point by Julian date, and additionally by the
    passed ARMC. """
    return _point(
        index=index,
        jd=jd,
        lat=lat,
        lon=None,
        house_system=house_system,
        pars_fortuna_formula=pars_fortuna_formula,
        armc=armc,
        armc_obliquity=obliquity
    )


def _objects(object_list: tuple, jd: float, lat: float, lon: float, house_system: int, pars_fortuna_formula: int, armc: float, armc_obliquity: float) -> dict:
    """ Function for items() and armc_items(). """
    objects = {}

    for index in object_list:
        objects[index] = _get(index, jd, lat, lon, house_system, pars_fortuna_formula, armc, armc_obliquity)

    return objects


def _get(index: int | str, jd: float, lat: float, lon: float, house_system: int, pars_fortuna_formula: int, armc: float, armc_obliquity: float) -> dict:
    """ Function for get() and armc_get(). """
    if armc is not None and armc_obliquity is None:
        armc_obliquity = obliquity(jd)

    if isinstance(index, int):
        if index < chart.TYPE_MULTIPLIER:
            return asteroid(index, jd)

        if index == chart.ANGLE:
            return _angle(ALL, jd, lat, lon, house_system, armc, armc_obliquity)

        if index == chart.HOUSE:
            return _house(ALL, jd, lat, lon, house_system, armc, armc_obliquity)

        match _type(index):
            case chart.ANGLE:
                return _angle(index, jd, lat, lon, house_system, armc, armc_obliquity)
            case chart.HOUSE:
                return _house(index, jd, lat, lon, house_system, armc, armc_obliquity)
            case chart.POINT:
                return _point(index, jd, lat, lon, house_system, pars_fortuna_formula, armc, armc_obliquity)
            case chart.ECLIPSE:
                return eclipse(index, jd)
            case (chart.ASTEROID|chart.PLANET):
                return planet(index, jd)
    else:
        return fixed_star(index, jd)


def _angle(index: int, jd: float, lat: float, lon: float, house_system: int, armc: float, armc_obliquity: float) -> dict:
    """ Function for angle() and armc_angle(). """
    if armc is not None:
        angles = _angles_houses_vertex_armc(armc, lat, armc_obliquity, house_system)['angles']
    else:
        angles = _angles_houses_vertex(jd, lat, lon, house_system)['angles']

    if index == ALL:
        return angles

    if index in angles:
        return angles[index]

    return None


def _house(index: int, jd: float, lat: float, lon: float, house_system: int, armc: float, armc_obliquity: float) -> dict:
    """ Function for house() and armc_house(). """
    if armc is not None:
        houses = _angles_houses_vertex_armc(armc, lat, armc_obliquity, house_system)['houses']
    else:
        houses = _angles_houses_vertex(jd, lat, lon, house_system)['houses']

    if index == ALL:
        return houses

    if index in houses:
        return houses[index]

    return None


def _point(index: int, jd: float, lat: float, lon: float, house_system: int, pars_fortuna_formula: int, armc: float, armc_obliquity: float) -> dict:
    """ Function for point() and armc_point(). """
    if index == chart.VERTEX:
        if armc is not None:
            return _angles_houses_vertex_armc(armc, lat, armc_obliquity, house_system)['vertex']
        else:
            return _angles_houses_vertex(jd, lat, lon, house_system)['vertex']

    if index == chart.SYZYGY:
        return _syzygy(jd)

    if index == chart.PARS_FORTUNA:
        return _pars_fortuna(jd, lat, lon, pars_fortuna_formula, armc, armc_obliquity)

    return _swisseph_point(index, jd)


@cache
def planet(index: int, jd: float) -> dict:
    """ Returns a pyswisseph object by Julian date. Can be used to
    return the six major asteroids supported by pyswisseph without using
    a separate ephemeris file. """
    ec_res = swe.calc_ut(jd, _SWE[index])[0]
    eq_res = swe.cotrans((ec_res[0], ec_res[1], ec_res[2]), -obliquity(jd))
    asteroid = _type(index) == chart.ASTEROID

    return {
        'index': index,
        'type': chart.ASTEROID if asteroid else chart.PLANET,
        'name': names.ASTEROIDS[index] if asteroid else names.PLANETS[index],
        'lon': ec_res[0],
        'lat': ec_res[1],
        'dist': ec_res[2],
        'speed': ec_res[3],
        'dec': eq_res[1],
    }


@cache
def asteroid(index: int, jd: float) -> dict:
    """ Returns an asteroid by Julian date and pyswisseph index
    from an external asteroid's ephmeris file as specified
    in the setup module. """
    if _type(index) == chart.ASTEROID:
        return planet(index, jd)

    swe_index = index + swe.AST_OFFSET
    name = swe.get_planet_name(swe_index)

    ec_res = swe.calc_ut(jd, swe_index)[0]
    eq_res = swe.cotrans((ec_res[0], ec_res[1], ec_res[2]), -obliquity(jd))

    return {
        'index': index,
        'type': chart.ASTEROID,
        'name': name,
        'lon': ec_res[0],
        'lat': ec_res[1],
        'dist': ec_res[2],
        'speed': ec_res[3],
        'dec': eq_res[1],
    }


@cache
def fixed_star(name: str, jd: float) -> dict:
    """ Returns a fixed star by Julian date and name. """
    res, stnam = swe.fixstar2_ut(name, jd)[:2]
    name = stnam.partition(',')[0]

    return {
        'index': name,
        'type': chart.FIXED_STAR,
        'name': name,
        'lon': res[0],
        'lat': res[1],
        'dist': res[2],
        'speed': res[3],
    }


@cache
def eclipse(index: int, jd: float) -> dict:
    """ Returns a calculated object based on the moon's or sun's position
    during a pre or post-natal lunar or solar eclipse. The declination
    value is based on the natal date. """
    match index:
        case chart.PRE_NATAL_SOLAR_ECLIPSE:
            eclipse_type, eclipse_jd = find.previous_solar_eclipse(jd)
            ec_res = swe.calc_ut(eclipse_jd, swe.SUN)[0]
        case chart.PRE_NATAL_LUNAR_ECLIPSE:
            eclipse_type, eclipse_jd = find.previous_lunar_eclipse(jd)
            ec_res = swe.calc_ut(eclipse_jd, swe.MOON)[0]
        case chart.POST_NATAL_SOLAR_ECLIPSE:
            eclipse_type, eclipse_jd = find.next_solar_eclipse(jd)
            ec_res = swe.calc_ut(eclipse_jd, swe.SUN)[0]
        case chart.POST_NATAL_LUNAR_ECLIPSE:
            eclipse_type, eclipse_jd = find.next_lunar_eclipse(jd)
            ec_res = swe.calc_ut(eclipse_jd, swe.MOON)[0]

    eq_res = swe.cotrans((ec_res[0], ec_res[1], ec_res[2]), -obliquity(jd))

    return {
        'index': index,
        'type': chart.ECLIPSE,
        'name': names.ECLIPSES[index],
        'eclipse_type': eclipse_type,
        'jd': eclipse_jd,
        'lon': ec_res[0],
        'lat': ec_res[1],
        'dist': ec_res[2],
        'speed': 0.0,
        'dec': eq_res[1],
    }


@cache
def moon_phase(jd: float) -> int:
    """ Returns the moon phase at the given Julian date. """
    sun = planet(chart.SUN, jd)
    moon = planet(chart.MOON, jd)
    return calculate.moon_phase(sun, moon)


@cache
def obliquity(jd: float, mean: bool = False) -> float:
    """ Returns the earth's true or mean obliquity at the
    given Julian date. """
    ecl_nut = swe.calc_ut(jd, swe.ECL_NUT)[0]
    return ecl_nut[1] if mean else ecl_nut[0]


def deltat(jd: float, seconds: bool = False) -> float:
    """ Return the Delta-T value of the passed Julian date. Optionally it
    will return this value in seconds. """
    return swe.deltat(jd) if not seconds else swe.deltat(jd) * 24 * 3600


def is_daytime(jd: float, lat: float, lon: float) -> bool:
    """ Returns whether the sun is above the horizon line at the time and
    place specified. """
    return _is_daytime(
        jd=jd,
        lat=lat,
        lon=lon,
        armc=None,
        armc_obliquity=None
    )


def armc_is_daytime(jd: float, armc: float, lat: float, obliquity: float) -> bool:
    """ Returns whether the sun is above the horizon line at the time and
    place specified, as calculated by the passed ARMC. """
    return _is_daytime(
        jd=jd,
        lat=lat,
        lon=None,
        armc=armc,
        armc_obliquity=obliquity
    )


@cache
def _is_daytime(jd: float, lat: float, lon: float, armc: float, armc_obliquity: float) -> bool:
    """ Function for is_daytime() and armc_is_daytime(). """
    sun = planet(chart.SUN, jd)
    asc = _angle(chart.ASC, jd, lat, lon, chart.PLACIDUS, armc, armc_obliquity)
    return calculate.is_daytime(sun, asc)


@cache
def _angles_houses_vertex(jd: float, lat: float, lon: float, house_system: int) -> dict:
    """ Returns ecliptic longitudes for the houses, main angles,
    and the vertex, along with their speeds. Based on Julian
    date and lat / lon coordinates. """
    return _angles_houses_vertex_from_swe(obliquity(jd), *swe.houses_ex2(jd, lat, lon, _SWE[house_system]))


@cache
def _angles_houses_vertex_armc(armc: float, lat: float, obliquity: float, house_system: int) -> dict:
    """ Returns ecliptic longitudes for the houses, main angles,
    and the vertex, along with their speeds. Based on ARMC, latitude
    and ecliptic obliquity. """
    return _angles_houses_vertex_from_swe(obliquity, *swe.houses_armc_ex2(armc, lat, obliquity, _SWE[house_system]))


def _angles_houses_vertex_from_swe(obliquity: float, cusps: tuple, ascmc: tuple, cuspsspeed: tuple, ascmcspeed: tuple) -> dict:
    """ Get houses, angles & vertex direct from pyswisseph. """
    angles = {}

    for i in (chart.ASC, chart.MC, chart.ARMC):
        lon = ascmc[_SWE[i]]
        speed = ascmcspeed[_SWE[i]]
        dec = swe.cotrans((lon, 0, 0), -obliquity)[1]

        angles[i] = {
            'index': i,
            'type': chart.ANGLE,
            'name': names.ANGLES[i],
            'lon': lon,
            'speed': speed,
            'dec': dec,
        }

        if i in (chart.ASC, chart.MC):
            index = chart.DESC if i == chart.ASC else chart.IC

            angles[index] = {
                'index': index,
                'type': chart.ANGLE,
                'name': names.ANGLES[index],
                'lon': swe.degnorm(lon - 180),
                'speed': speed,
                'dec': dec * -1,
            }

    houses = {}

    for i, lon in enumerate(cusps, start=1):
        index = chart.HOUSE + i
        size = swe.difdeg2n(cusps[i if i < 12 else 0], lon)
        speed = cuspsspeed[i-1]
        dec = swe.cotrans((lon, 0, 0), -obliquity)[1]

        houses[index] = {
            'index': index,
            'type': chart.HOUSE,
            'name': names.HOUSES[index],
            'number': i,
            'lon': lon,
            'size': size,
            'speed': speed,
            'dec': dec,
        }

    vertex_lon = ascmc[_SWE[chart.VERTEX]]
    vertex_speed = ascmcspeed[_SWE[chart.VERTEX]]
    vertex_dec = swe.cotrans((vertex_lon, 0, 0), -obliquity)[1]

    vertex = {
        'index': chart.VERTEX,
        'type': chart.POINT,
        'name': names.POINTS[chart.VERTEX],
        'lon': vertex_lon,
        'speed': vertex_speed,
        'dec': vertex_dec,
    }

    return {
        'angles': angles,
        'houses': houses,
        'vertex': vertex,
    }


@cache
def _syzygy(jd: float) -> dict:
    """ Calculate prenatal full/new moon - this can potentially
    be an expensive calculation so should be cached. """
    sun = planet(chart.SUN, jd)
    moon = planet(chart.MOON, jd)
    distance = swe.difdeg2n(moon['lon'], sun['lon'])
    syzygy_jd = find.previous_new_moon(jd) if distance > 0 else find.previous_full_moon(jd)
    syzygy_moon = planet(chart.MOON, syzygy_jd)

    return {
        'index': chart.SYZYGY,
        'type': chart.POINT,
        'name': names.POINTS[chart.SYZYGY],
        'lon': syzygy_moon['lon'],
        'lat': syzygy_moon['lat'],
        'speed': syzygy_moon['speed'],
        'dec': syzygy_moon['dec'],
    }


@cache
def _pars_fortuna(jd: float, lat: float, lon: float, formula: int, armc: float = None, armc_obliquity: float = None) -> dict:
    """ Calculate Part of Fortune - this relies on the ascendant which will be
    consistent across all supported systems, so it is safe to pass Placidus as
    the default. """
    sun = planet(chart.SUN, jd)
    moon = planet(chart.MOON, jd)
    asc = angle(chart.ASC, jd, lat, lon, chart.PLACIDUS) if armc is None else armc_angle(chart.ASC, armc, lat, armc_obliquity, chart.PLACIDUS)
    lon = calculate.pars_fortuna_longitude(sun, moon, asc, formula)
    dec = swe.cotrans((lon, 0, 0), -obliquity(jd))[1]

    return {
        'index': chart.PARS_FORTUNA,
        'type': chart.POINT,
        'name': names.POINTS[chart.PARS_FORTUNA],
        'lon': lon,
        'lat': 0.0,
        'speed': 0.0,
        'dec': dec,
    }


@cache
def _swisseph_point(index: int, jd: float) -> dict:
    """ Pull any remaining non-calculated points straight from pyswisseph. """
    res = swe.calc_ut(jd, _SWE[index])[0]
    lon = res[0] if index not in (chart.SOUTH_NODE, chart.TRUE_SOUTH_NODE) else swe.degnorm(res[0] - 180)
    lat = res[1] if index not in (chart.NORTH_NODE, chart.TRUE_NORTH_NODE, chart.SOUTH_NODE, chart.TRUE_SOUTH_NODE) else 0.0
    speed = res[3]
    dec = swe.cotrans((lon, lat, 0), -obliquity(jd))[1]

    return {
        'index': index,
        'type': chart.POINT,
        'name': names.POINTS[index],
        'lon': lon,
        'lat': lat,
        'speed': speed,
        'dec': dec,
    }


def _type(index: int) -> int:
    """ Return the type index of a given object's index. """
    return round(index, -2)



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Tests the main chart classes. Most of the underlying functionality
    comes from the submodules which have their own tests, so this file
    focuses on testing that the classes are working with and returning
    the correct charts, rather than re-testing the minutiae of each of
    their data. If all spot-checks for a given chart return true, then
    we can safely assume we are looking at the correct chart and trust
    that the rest of the underlying data is correct based on the tests
    for the submodules.

"""


from datetime import datetime
from zoneinfo import ZoneInfo

from pytest import fixture

from immanuel import charts
from immanuel.const import calc, chart, names
from immanuel.setup import settings
from immanuel.tools import convert


@fixture
def dob():
    return '2000-01-01 10:00'

@fixture
def partner_dob():
    return '2001-02-16 06:00'

@fixture
def lat():
    return '32N43.0'

@fixture
def partner_lat():
    return '38N35.0'

@fixture
def lon():
    return '117W9.0'

@fixture
def partner_lon():
    return '121W30.0'

@fixture
def native(dob, lat, lon):
    return charts.Subject(dob, lat, lon)

@fixture
def partner(partner_dob, partner_lat, partner_lon):
    return charts.Subject(partner_dob, partner_lat, partner_lon)

@fixture
def solar_return_year():
    return 2030

@fixture
def pdt():
    # We compare against astro.com which assumes UTC for progressions date
    # so we knock 8 hours off midnight 2023-06-21 to account for Pacific Time
    return '2025-06-20 17:00:00'


def test_subject(dob, lat, lon, native):
    date_time = datetime.fromisoformat(f'{dob} -08:00')
    latitude, longitude = (convert.string_to_dec(v) for v in (lat, lon))
    assert native.date_time.year == date_time.year
    assert native.date_time.month == date_time.month
    assert native.date_time.day == date_time.day
    assert native.date_time.hour == date_time.hour
    assert native.date_time.minute == date_time.minute
    assert native.date_time.second == date_time.second
    assert native.latitude == latitude
    assert native.longitude == longitude


def test_natal(native, lat, lon):
    natal_chart = charts.Natal(native)

    assert natal_chart.type == names.CHART_TYPES[chart.NATAL]

    assert round(natal_chart.native.date_time.julian + natal_chart.native.date_time.deltat, 6) == 2451545.250739
    assert natal_chart.native.date_time.timezone == 'PST'

    assert natal_chart.native.coordinates.latitude.formatted == lat
    assert natal_chart.native.coordinates.longitude.formatted == lon

    assert natal_chart.house_system == names.HOUSE_SYSTEMS[settings.house_system]
    assert natal_chart.shape == names.CHART_SHAPES[calc.BOWL]
    assert natal_chart.diurnal == True
    assert natal_chart.moon_phase.third_quarter == True

    # Spot-check for correct object positions against astro.com
    assert natal_chart.objects[chart.SUN].name == names.PLANETS[chart.SUN]
    assert natal_chart.objects[chart.SUN].sign.name == names.SIGNS[chart.CAPRICORN]
    assert natal_chart.objects[chart.SUN].sign_longitude.formatted == '10°37\'26"'

    assert natal_chart.objects[chart.MOON].name == names.PLANETS[chart.MOON]
    assert natal_chart.objects[chart.MOON].sign.name == names.SIGNS[chart.SCORPIO]
    assert natal_chart.objects[chart.MOON].sign_longitude.formatted == '16°19\'29"'

    assert natal_chart.objects[chart.PARS_FORTUNA].name == names.POINTS[chart.PARS_FORTUNA]
    assert natal_chart.objects[chart.PARS_FORTUNA].sign.name == names.SIGNS[chart.CAPRICORN]
    assert natal_chart.objects[chart.PARS_FORTUNA].sign_longitude.formatted == '11°18\'41"'

    # Spot-check for correct object data against astro.com & Astro Gold
    assert natal_chart.objects[chart.SATURN].movement.retrograde == True
    assert natal_chart.objects[chart.MARS].dignities.peregrine == True
    assert natal_chart.objects[chart.MARS].score == -5

    # Spot-check for correct angle positions against astro.com
    assert natal_chart.objects[chart.ASC].name == names.ANGLES[chart.ASC]
    assert natal_chart.objects[chart.ASC].sign.name == names.SIGNS[chart.PISCES]
    assert natal_chart.objects[chart.ASC].sign_longitude.formatted == '05°36\'38"'

    assert natal_chart.objects[chart.MC].name == names.ANGLES[chart.MC]
    assert natal_chart.objects[chart.MC].sign.name == names.SIGNS[chart.SAGITTARIUS]
    assert natal_chart.objects[chart.MC].sign_longitude.formatted == '14°50\'44"'

    # Spot-check for correct 2nd house position against astro.com
    assert natal_chart.houses[chart.HOUSE2].name == names.HOUSES[chart.HOUSE2]
    assert natal_chart.houses[chart.HOUSE2].sign.name == names.SIGNS[chart.ARIES]
    assert natal_chart.houses[chart.HOUSE2].sign_longitude.formatted == '17°59\'40"'

    # Spot-check for correct aspects against astro.com
    assert chart.SUN in natal_chart.aspects
    assert chart.MOON in natal_chart.aspects[chart.SUN]
    assert natal_chart.aspects[chart.SUN][chart.MOON].aspect == calc.SEXTILE

    assert chart.MOON in natal_chart.aspects
    assert chart.SATURN in natal_chart.aspects[chart.MOON]
    assert natal_chart.aspects[chart.MOON][chart.SATURN].aspect == calc.OPPOSITION

    # Spot-check for correct weightings against astro.com
    assert chart.JUPITER in natal_chart.weightings['elements'].fire
    assert chart.JUPITER in natal_chart.weightings['modalities'].cardinal
    assert chart.JUPITER in natal_chart.weightings['quadrants'].first


def test_solar_return(native, lat, lon, solar_return_year):
    solar_return_chart = charts.SolarReturn(native, solar_return_year)

    assert solar_return_chart.type == names.CHART_TYPES[chart.SOLAR_RETURN]

    assert round(solar_return_chart.native.date_time.julian + solar_return_chart.native.date_time.deltat, 6) == 2451545.250739
    assert solar_return_chart.native.date_time.timezone == 'PST'

    assert round(solar_return_chart.solar_return_date_time.julian + solar_return_chart.solar_return_date_time.deltat, 6) == 2462502.521823
    assert solar_return_chart.solar_return_date_time.timezone == 'PST'

    assert solar_return_chart.native.coordinates.latitude.formatted == lat
    assert solar_return_chart.native.coordinates.longitude.formatted == lon

    assert solar_return_chart.house_system == names.HOUSE_SYSTEMS[settings.house_system]
    assert solar_return_chart.shape == names.CHART_SHAPES[calc.LOCOMOTIVE]
    assert solar_return_chart.diurnal == True
    assert solar_return_chart.moon_phase.balsamic == True

    # Spot-check for correct object positions against astro.com
    assert solar_return_chart.objects[chart.SUN].name == names.PLANETS[chart.SUN]
    assert solar_return_chart.objects[chart.SUN].sign.name == names.SIGNS[chart.CAPRICORN]
    assert solar_return_chart.objects[chart.SUN].sign_longitude.formatted == '10°37\'26"'

    assert solar_return_chart.objects[chart.MOON].name == names.PLANETS[chart.MOON]
    assert solar_return_chart.objects[chart.MOON].sign.name == names.SIGNS[chart.SCORPIO]
    assert solar_return_chart.objects[chart.MOON].sign_longitude.formatted == '28°43\'43"'

    assert solar_return_chart.objects[chart.PARS_FORTUNA].name == names.POINTS[chart.PARS_FORTUNA]
    assert solar_return_chart.objects[chart.PARS_FORTUNA].sign.name == names.SIGNS[chart.TAURUS]
    assert solar_return_chart.objects[chart.PARS_FORTUNA].sign_longitude.formatted == '24°41\'28"'

    # Spot-check for correct object data against astro.com & Astro Gold
    assert solar_return_chart.objects[chart.SATURN].movement.retrograde == True
    assert solar_return_chart.objects[chart.MARS].dignities.peregrine == True
    assert solar_return_chart.objects[chart.MARS].score == -5

    # Spot-check for correct angle positions against astro.com
    assert solar_return_chart.objects[chart.ASC].name == names.ANGLES[chart.ASC]
    assert solar_return_chart.objects[chart.ASC].sign.name == names.SIGNS[chart.CANCER]
    assert solar_return_chart.objects[chart.ASC].sign_longitude.formatted == '06°35\'11"'

    assert solar_return_chart.objects[chart.MC].name == names.ANGLES[chart.MC]
    assert solar_return_chart.objects[chart.MC].sign.name == names.SIGNS[chart.PISCES]
    assert solar_return_chart.objects[chart.MC].sign_longitude.formatted == '20°21\'06"'

    # Spot-check for correct 2nd house position against astro.com
    assert solar_return_chart.houses[chart.HOUSE2].name == names.HOUSES[chart.HOUSE2]
    assert solar_return_chart.houses[chart.HOUSE2].sign.name == names.SIGNS[chart.CANCER]
    assert solar_return_chart.houses[chart.HOUSE2].sign_longitude.formatted == '28°17\'34"'

    # Spot-check for correct aspects against astro.com
    assert chart.SUN in solar_return_chart.aspects
    assert chart.SATURN in solar_return_chart.aspects[chart.SUN]
    assert solar_return_chart.aspects[chart.SUN][chart.SATURN].aspect == calc.TRINE

    assert chart.MOON in solar_return_chart.aspects
    assert chart.NEPTUNE in solar_return_chart.aspects[chart.MOON]
    assert solar_return_chart.aspects[chart.MOON][chart.NEPTUNE].aspect == calc.TRINE

    # Spot-check for correct weightings against astro.com
    assert chart.JUPITER in solar_return_chart.weightings['elements'].water
    assert chart.JUPITER in solar_return_chart.weightings['modalities'].fixed
    assert chart.JUPITER in solar_return_chart.weightings['quadrants'].second


def test_progressed(native, lat, lon, pdt):
    settings.mc_progression_method = calc.NAIBOD
    progressed_chart = charts.Progressed(native, pdt)

    assert progressed_chart.type == names.CHART_TYPES[chart.PROGRESSED]
    assert progressed_chart.progression_method == names.PROGRESSION_METHODS[settings.mc_progression_method]

    assert round(progressed_chart.native.date_time.julian + progressed_chart.native.date_time.deltat, 6) == 2451545.250739
    assert progressed_chart.native.date_time.timezone == 'PST'

    assert progressed_chart.progression_date_time.timezone == 'PDT'
    pdt_utc = progressed_chart.progression_date_time.datetime.astimezone(ZoneInfo('UTC'))
    assert (pdt_utc.year, pdt_utc.month, pdt_utc.day, pdt_utc.hour) == (2025, 6, 21, 0)

    # Progressed date tested against astro.com
    assert round(progressed_chart.progressed_date_time.julian + progressed_chart.progressed_date_time.deltat, 6) == 2451570.719456
    assert progressed_chart.progressed_date_time.timezone == 'PST'

    # Ensure coords have been converted back into correct string
    assert progressed_chart.native.coordinates.latitude.formatted == lat
    assert progressed_chart.native.coordinates.longitude.formatted == lon

    assert progressed_chart.house_system == names.HOUSE_SYSTEMS[settings.house_system]
    assert progressed_chart.shape == names.CHART_SHAPES[calc.LOCOMOTIVE]
    assert progressed_chart.diurnal == True
    assert progressed_chart.moon_phase.disseminating == True

    # Spot-check for correct object positions against astro.com
    assert progressed_chart.objects[chart.SUN].name == names.PLANETS[chart.SUN]
    assert progressed_chart.objects[chart.SUN].sign.name == names.SIGNS[chart.AQUARIUS]
    assert progressed_chart.objects[chart.SUN].sign_longitude.formatted == '06°33\'41"'

    assert progressed_chart.objects[chart.MOON].name == names.PLANETS[chart.MOON]
    assert progressed_chart.objects[chart.MOON].sign.name == names.SIGNS[chart.LIBRA]
    assert progressed_chart.objects[chart.MOON].sign_longitude.formatted == '23°50\'57"'

    assert progressed_chart.objects[chart.PARS_FORTUNA].name == names.POINTS[chart.PARS_FORTUNA]
    assert progressed_chart.objects[chart.PARS_FORTUNA].sign.name == names.SIGNS[chart.CAPRICORN]
    assert progressed_chart.objects[chart.PARS_FORTUNA].sign_longitude.formatted == '00°17\'45"'

    # Spot-check for correct object data against Astro Gold
    assert progressed_chart.objects[chart.SUN].dignities.detriment == True
    assert progressed_chart.objects[chart.SUN].dignities.peregrine == True
    assert progressed_chart.objects[chart.SUN].score == -10

    # Spot-check for correct angle positions against astro.com
    assert progressed_chart.objects[chart.ASC].name == names.ANGLES[chart.ASC]
    assert progressed_chart.objects[chart.ASC].sign.name == names.SIGNS[chart.ARIES]
    assert progressed_chart.objects[chart.ASC].sign_longitude.formatted == '13°00\'29"'

    assert progressed_chart.objects[chart.MC].name == names.ANGLES[chart.MC]
    assert progressed_chart.objects[chart.MC].sign.name == names.SIGNS[chart.CAPRICORN]
    assert progressed_chart.objects[chart.MC].sign_longitude.formatted == '07°57\'07"'

    # Spot-check for correct 2nd house position against astro.com
    assert progressed_chart.houses[chart.HOUSE2].name == names.HOUSES[chart.HOUSE2]
    assert progressed_chart.houses[chart.HOUSE2].sign.name == names.SIGNS[chart.TAURUS]
    assert progressed_chart.houses[chart.HOUSE2].sign_longitude.formatted == '18°52\'16"'

    # Spot-check for correct aspects against astro.com
    assert chart.SUN in progressed_chart.aspects
    assert chart.SATURN in progressed_chart.aspects[chart.SUN]
    assert progressed_chart.aspects[chart.SUN][chart.SATURN].aspect == calc.SQUARE

    assert chart.MOON in progressed_chart.aspects
    assert chart.URANUS in progressed_chart.aspects[chart.MOON]
    assert progressed_chart.aspects[chart.MOON][chart.URANUS].aspect == calc.TRINE

    # Spot-check for correct weightings against astro.com
    assert chart.VENUS in progressed_chart.weightings['elements'].earth
    assert chart.VENUS in progressed_chart.weightings['modalities'].cardinal
    assert chart.VENUS in progressed_chart.weightings['quadrants'].third


def test_composite(native, lat, lon, partner, partner_lat, partner_lon):
    composite_chart = charts.Composite(native, partner)

    assert composite_chart.type == names.CHART_TYPES[chart.COMPOSITE]

    assert round(composite_chart.native.date_time.julian + composite_chart.native.date_time.deltat, 6) == 2451545.250739
    assert composite_chart.native.date_time.timezone == 'PST'

    assert round(composite_chart.partner.date_time.julian + composite_chart.partner.date_time.deltat, 6) == 2451957.084075
    assert composite_chart.partner.date_time.timezone == 'PST'

    assert composite_chart.native.coordinates.latitude.formatted == lat
    assert composite_chart.native.coordinates.longitude.formatted == lon

    assert composite_chart.partner.coordinates.latitude.formatted == partner_lat
    assert composite_chart.partner.coordinates.longitude.formatted == partner_lon

    assert composite_chart.house_system == names.HOUSE_SYSTEMS[settings.house_system]
    assert composite_chart.diurnal == True
    assert composite_chart.moon_phase.third_quarter == True

    # Spot-check for correct object positions against astro.com
    assert composite_chart.objects[chart.SUN].name == names.PLANETS[chart.SUN]
    assert composite_chart.objects[chart.SUN].sign.name == names.SIGNS[chart.AQUARIUS]
    assert composite_chart.objects[chart.SUN].sign_longitude.formatted == '04°17\'35"'

    assert composite_chart.objects[chart.MOON].name == names.PLANETS[chart.MOON]
    assert composite_chart.objects[chart.MOON].sign.name == names.SIGNS[chart.SAGITTARIUS]
    assert composite_chart.objects[chart.MOON].sign_longitude.formatted == '00°22\'35"'

    assert composite_chart.objects[chart.PARS_FORTUNA].name == names.POINTS[chart.PARS_FORTUNA]
    assert composite_chart.objects[chart.PARS_FORTUNA].sign.name == names.SIGNS[chart.PISCES]
    assert composite_chart.objects[chart.PARS_FORTUNA].sign_longitude.formatted == '01°03\'58"'

    # Spot-check for correct object data against Astro Gold
    assert composite_chart.objects[chart.SUN].dignities.detriment == True
    assert composite_chart.objects[chart.SUN].dignities.peregrine == True
    assert composite_chart.objects[chart.SUN].score == -10

    # Spot-check for correct angle positions against astro.com
    assert composite_chart.objects[chart.ASC].name == names.ANGLES[chart.ASC]
    assert composite_chart.objects[chart.ASC].sign.name == names.SIGNS[chart.AQUARIUS]
    assert composite_chart.objects[chart.ASC].sign_longitude.formatted == '21°26\'55"'

    assert composite_chart.objects[chart.MC].name == names.ANGLES[chart.MC]
    assert composite_chart.objects[chart.MC].sign.name == names.SIGNS[chart.SAGITTARIUS]
    assert composite_chart.objects[chart.MC].sign_longitude.formatted == '06°07\'28"'

    # Spot-check for correct 2nd house position against astro.com
    assert composite_chart.houses[chart.HOUSE2].name == names.HOUSES[chart.HOUSE2]
    assert composite_chart.houses[chart.HOUSE2].sign.name == names.SIGNS[chart.ARIES]
    assert composite_chart.houses[chart.HOUSE2].sign_longitude.formatted == '05°32\'57"'

    # Spot-check for correct aspects against astro.com
    assert chart.SUN in composite_chart.aspects
    assert chart.MOON in composite_chart.aspects[chart.SUN]
    assert composite_chart.aspects[chart.SUN][chart.MOON].aspect == calc.SEXTILE

    assert chart.MOON in composite_chart.aspects
    assert chart.VENUS in composite_chart.aspects[chart.MOON]
    assert composite_chart.aspects[chart.MOON][chart.VENUS].aspect == calc.SEXTILE

    # Spot-check for correct weightings against astro.com
    assert chart.JUPITER in composite_chart.weightings['elements'].earth
    assert chart.JUPITER in composite_chart.weightings['modalities'].fixed
    assert chart.JUPITER in composite_chart.weightings['quadrants'].first


def test_transits(lat, lon):
    transits_chart = charts.Transits(lat, lon)

    assert transits_chart.type == names.CHART_TYPES[chart.TRANSITS]

    assert transits_chart.native.coordinates.latitude.formatted == lat
    assert transits_chart.native.coordinates.longitude.formatted == lon

    assert transits_chart.house_system == names.HOUSE_SYSTEMS[settings.house_system]

    default_transits_chart = charts.Transits()

    assert default_transits_chart.native.coordinates.latitude.raw == settings.default_latitude
    assert default_transits_chart.native.coordinates.longitude.raw == settings.default_longitude


def test_synastry(native, partner):
    partner_chart = charts.Natal(partner)
    native_chart = charts.Natal(native, aspects_to=partner_chart)

    # Spot-check for correct aspects against astro.com
    assert chart.SUN in native_chart.aspects
    assert chart.VENUS in native_chart.aspects[chart.SUN]
    assert native_chart.aspects[chart.SUN][chart.VENUS].aspect == calc.SQUARE

    assert chart.MOON in native_chart.aspects
    assert chart.MERCURY in native_chart.aspects[chart.MOON]
    assert native_chart.aspects[chart.MOON][chart.MERCURY].aspect == calc.SQUARE



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    The ephemeris module's figures are tested against output
    from astro.com with default Placidus house system.
    Additional data courtesy of websites cited in test
    function comments.

"""

import os
from datetime import datetime

from pytest import approx, fixture

from immanuel import setup
from immanuel.const import calc, chart
from immanuel.tools import convert, date, ephemeris, position


@fixture
def coords():
    # San Diego coords as used by astro.com
    return [convert.string_to_dec(v) for v in ('32n43', '117w09')]

@fixture
def jd(coords):
    return date.to_jd('2000-01-01 10:00', *coords)

@fixture
def armc():
    # ARMC longitude on the above Julian date
    return 253.55348499294269

@fixture
def all_angles():
    return (
        chart.ASC,
        chart.DESC,
        chart.MC,
        chart.IC,
        chart.ARMC,
    )

@fixture
def all_houses():
    return (
        chart.HOUSE1,
        chart.HOUSE2,
        chart.HOUSE3,
        chart.HOUSE4,
        chart.HOUSE5,
        chart.HOUSE6,
        chart.HOUSE7,
        chart.HOUSE8,
        chart.HOUSE9,
        chart.HOUSE10,
        chart.HOUSE11,
        chart.HOUSE12,
    )

@fixture
def all_points():
    return (
        chart.NORTH_NODE,
        chart.SOUTH_NODE,
        chart.TRUE_NORTH_NODE,
        chart.TRUE_SOUTH_NODE,
        chart.VERTEX,
        chart.LILITH,
        chart.TRUE_LILITH,
        chart.SYZYGY,
        chart.PARS_FORTUNA,
    )

@fixture
def all_planets():
    return (
        chart.SUN,
        chart.MOON,
        chart.MERCURY,
        chart.VENUS,
        chart.MARS,
        chart.JUPITER,
        chart.SATURN,
        chart.URANUS,
        chart.NEPTUNE,
        chart.PLUTO,
    )

@fixture
def all_asteroids():
    return (
        chart.CHIRON,
        chart.PHOLUS,
        chart.CERES,
        chart.PALLAS,
        chart.JUNO,
        chart.VESTA,
        1181,
    )

@fixture
def astro():
    """ Results copied from astro.com chart table. We spot-check
    chart objects by picking one of each type. """
    return {
        # angle
        'asc': {
            'lon': '05°36\'38"',
            # This is the only figure disagreeing with astro.com (~1 arcsec) and nobody knows why
            # 'dec': '-09°27\'13"',
        },
        # house
        'house_2': {
            'lon': '17°59\'40"',
            'dec': '07°03\'29"',
        },
        # planet
        'sun': {
            'lon': '10°37\'26"',
            'lat': '00°00\'01"',
            'speed': '01°01\'10"',
            'dec': '-23°00\'45"',
        },
        # point
        'pof': {
            'lon': '11°18\'41"',
            'dec': '-22°57\'22"',
        },
        # default asteroid
        'juno': {
            'lon': '08°05\'21"',
            'lat': '09°26\'57"',
            'speed': '00°22\'21"',
            'dec': '-13°45\'30"',
        },
        # external asteroid
        'lilith': {
            'lon': '18°16\'47"',
            'lat': '04°49\'07"',
            'speed': '00°24\'37"',
            'dec': '-00°11\'50"',
        },
        # fixed star
        'antares': {
            'lon': '09°45\'12"',
            'lat': '-04°34\'11"',
        },
        # eclipses
        'pre_natal_solar_eclipse': {
            'lon': '18°20\'59"',
            'lat': '00°00\'00"',
            'speed': '00°00\'00"',
            'dec': '15°19\'40"',
            'eclipse_type': chart.TOTAL,
            'date': '11 August',
        },
        'pre_natal_lunar_eclipse': {
            'lon': '05°02\'21"',
            'lat': '00°43\'35"',
            'speed': '00°00\'00"',
            'dec': '-18°18\'03"',
            'eclipse_type': chart.PARTIAL,
            'date': '28 July',
        },
        'post_natal_solar_eclipse': {
            'lon': '16°01\'14"',
            'lat': '00°00\'00"',
            'speed': '00°00\'00"',
            'dec': '-16°02\'00"',
            'eclipse_type': chart.PARTIAL,
            'date': '05 February',
        },
        'post_natal_lunar_eclipse': {
            'lon': '00°28\'04"',
            'lat': '-00°17\'53"',
            'speed': '00°00\'00"',
            'dec': '19°45\'29"',
            'eclipse_type': chart.TOTAL,
            'date': '20 January',
        },
    }


""" These tests simply check the correct chart objects are being
returned. Data is checked separately afterwards. """
def test_objects(jd, coords):
    chart_objects = (chart.SUN, chart.MOON, chart.PARS_FORTUNA, chart.SYZYGY, chart.NORTH_NODE, chart.ASC)
    objects = ephemeris.objects(chart_objects, jd, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
    assert tuple(objects.keys()) == chart_objects


def test_armc_objects(jd, coords, armc):
    chart_objects = (chart.SUN, chart.MOON, chart.PARS_FORTUNA, chart.SYZYGY, chart.NORTH_NODE, chart.ASC)
    objects = ephemeris.armc_objects(chart_objects, jd, armc, *coords, None, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
    assert tuple(objects.keys()) == chart_objects


def test_get(jd, coords):
    setup.add_filepath(os.path.dirname(__file__))
    assert ephemeris.get(chart.ASC, jd, *coords, chart.PLACIDUS)['index'] == chart.ASC
    assert ephemeris.get(chart.HOUSE2, jd, *coords, chart.PLACIDUS)['index'] == chart.HOUSE2
    assert ephemeris.get(chart.SUN, jd)['index'] == chart.SUN
    assert ephemeris.get(chart.PARS_FORTUNA, jd, *coords, pars_fortuna_formula=calc.DAY_NIGHT_FORMULA)['index'] == chart.PARS_FORTUNA
    assert ephemeris.get(chart.JUNO, jd)['index'] == chart.JUNO   # Included with planets
    lilith = ephemeris.get(1181, jd)                              # From external file
    antares = ephemeris.get('Antares', jd)
    assert lilith['index'] == 1181 and lilith['type'] == chart.ASTEROID
    assert antares['index'] == 'Antares' and antares['type'] == chart.FIXED_STAR


def test_armc_get(jd, coords, armc):
    setup.add_filepath(os.path.dirname(__file__))
    assert ephemeris.armc_get(chart.ASC, jd, armc, coords[0], house_system=chart.PLACIDUS)['index'] == chart.ASC
    assert ephemeris.armc_get(chart.HOUSE2, jd, armc, coords[0], house_system=chart.PLACIDUS)['index'] == chart.HOUSE2
    assert ephemeris.armc_get(chart.PARS_FORTUNA, jd, armc, coords[0], pars_fortuna_formula=calc.DAY_NIGHT_FORMULA)['index'] == chart.PARS_FORTUNA


def test_get_angles(jd, coords, all_angles):
    angles = ephemeris.get(chart.ANGLE, jd, *coords, chart.PLACIDUS)
    assert sorted(all_angles) == sorted(angles)


def test_armc_get_angles(jd, coords, armc, all_angles):
    angles = ephemeris.armc_get(chart.ANGLE, jd, armc, *coords, ephemeris.obliquity(jd), chart.PLACIDUS)
    assert sorted(all_angles) == sorted(angles)


def test_get_houses(jd, coords, all_houses):
    houses = ephemeris.get(chart.HOUSE, jd, *coords, chart.PLACIDUS)
    assert sorted(all_houses) == sorted(houses)


def test_armc_get_houses(jd, coords, armc, all_houses):
    houses = ephemeris.armc_get(chart.HOUSE, jd, armc, *coords, ephemeris.obliquity(jd), chart.PLACIDUS)
    assert sorted(all_houses) == sorted(houses)


def test_angles(jd, coords, all_angles):
    angles = ephemeris.angles(jd, *coords, chart.PLACIDUS)
    assert sorted(all_angles) == sorted(angles)


def test_armc_angles(jd, coords, armc, all_angles):
    angles = ephemeris.armc_angles(armc, coords[0], ephemeris.obliquity(jd), chart.PLACIDUS)
    assert sorted(all_angles) == sorted(angles)


def test_angle(jd, coords, all_angles):
    for index in all_angles:
        angle = ephemeris.angle(index, jd, *coords, chart.PLACIDUS)
        assert angle['index'] == index and angle['type'] == chart.ANGLE

    assert ephemeris.angle(ephemeris.ALL, jd, *coords, chart.PLACIDUS) == ephemeris.angles(jd, *coords, chart.PLACIDUS)


def test_armc_angle(jd, coords, armc, all_angles):
    obliquity = ephemeris.obliquity(jd)

    for index in all_angles:
        angle = ephemeris.armc_angle(index, armc, coords[0], obliquity, chart.PLACIDUS)
        assert angle['index'] == index and angle['type'] == chart.ANGLE

    assert ephemeris.armc_angle(ephemeris.ALL, armc, coords[0], obliquity, chart.PLACIDUS) == ephemeris.armc_angles(armc, coords[0], obliquity, chart.PLACIDUS)


def test_houses(jd, coords, all_houses):
    houses = ephemeris.houses(jd, *coords, chart.PLACIDUS)
    assert sorted(all_houses) == sorted(houses)


def test_armc_houses(jd, coords, armc, all_houses):
    houses = ephemeris.armc_houses(armc, coords[0], ephemeris.obliquity(jd), chart.PLACIDUS)
    assert sorted(all_houses) == sorted(houses)


def test_house(jd, coords, all_houses):
    for index in all_houses:
        house = ephemeris.house(index, jd, *coords, chart.PLACIDUS)
        assert house['index'] == index and house['type'] == chart.HOUSE

    assert ephemeris.house(ephemeris.ALL, jd, *coords, chart.PLACIDUS) == ephemeris.houses(jd, *coords, chart.PLACIDUS)


def test_armc_house(jd, coords, armc, all_houses):
    obliquity = ephemeris.obliquity(jd)

    for index in all_houses:
        house = ephemeris.armc_house(index, armc, coords[0], obliquity, chart.PLACIDUS)
        assert house['index'] == index and house['type'] == chart.HOUSE

    assert ephemeris.armc_house(ephemeris.ALL, armc, coords[0], obliquity, chart.PLACIDUS) == ephemeris.armc_houses(armc, coords[0], obliquity, chart.PLACIDUS)


def test_point(jd, coords, all_points):
    for index in all_points:
        point = ephemeris.point(index, jd, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
        assert point['index'] == index and point['type'] == chart.POINT


def test_armc_point(jd, coords, armc, all_points):
    for index in all_points:
        point = ephemeris.armc_point(index, jd, armc, coords[0], ephemeris.obliquity(jd), chart.PLACIDUS, calc.DAY_NIGHT_FORMULA)
        assert point['index'] == index and point['type'] == chart.POINT


def test_planet(jd, all_planets):
    for index in all_planets:
        planet = ephemeris.planet(index, jd)
        assert planet['index'] == index and planet['type'] == chart.PLANET


def test_asteroid(jd, all_asteroids):
    setup.add_filepath(os.path.dirname(__file__))

    for index in all_asteroids:
        asteroid = ephemeris.asteroid(index, jd)
        assert asteroid['index'] == index and asteroid['type'] == chart.ASTEROID


def test_fixed_star(jd):
    # So many fixed stars we just test one
    fixed_star = ephemeris.fixed_star('Antares', jd)
    assert fixed_star['index'] == 'Antares' and fixed_star['type'] == chart.FIXED_STAR


def test_eclipse(jd):
    pre_solar = ephemeris.eclipse(chart.PRE_NATAL_SOLAR_ECLIPSE, jd)
    assert pre_solar['type'] == chart.ECLIPSE
    assert pre_solar['index'] == chart.PRE_NATAL_SOLAR_ECLIPSE

    pre_lunar = ephemeris.eclipse(chart.PRE_NATAL_LUNAR_ECLIPSE, jd)
    assert pre_lunar['type'] == chart.ECLIPSE
    assert pre_lunar['index'] == chart.PRE_NATAL_LUNAR_ECLIPSE

    post_solar = ephemeris.eclipse(chart.POST_NATAL_SOLAR_ECLIPSE, jd)
    assert post_solar['type'] == chart.ECLIPSE
    assert post_solar['index'] == chart.POST_NATAL_SOLAR_ECLIPSE

    post_lunar = ephemeris.eclipse(chart.POST_NATAL_LUNAR_ECLIPSE, jd)
    assert post_lunar['type'] == chart.ECLIPSE
    assert post_lunar['index'] == chart.POST_NATAL_LUNAR_ECLIPSE


""" Now we are satisfied the correct chart objects are being returned,
we can test the accuracy of the module's data. """
def test_get_data(coords, jd, astro):
    setup.add_filepath(os.path.dirname(__file__))

    data = {
        'asc': ephemeris.angle(chart.ASC, jd, *coords, chart.PLACIDUS),
        'house_2': ephemeris.house(chart.HOUSE2, jd, *coords, chart.PLACIDUS),
        'sun': ephemeris.planet(chart.SUN, jd),
        'pof': ephemeris.point(chart.PARS_FORTUNA, jd, *coords, pars_fortuna_formula=calc.DAY_NIGHT_FORMULA),
        'juno': ephemeris.asteroid(chart.JUNO, jd),    # Included with planets
        'lilith': ephemeris.asteroid(1181, jd),        # From external file
        'antares': ephemeris.fixed_star('Antares', jd),
        'pre_natal_solar_eclipse': ephemeris.eclipse(chart.PRE_NATAL_SOLAR_ECLIPSE, jd),
        'pre_natal_lunar_eclipse': ephemeris.eclipse(chart.PRE_NATAL_LUNAR_ECLIPSE, jd),
        'post_natal_solar_eclipse': ephemeris.eclipse(chart.POST_NATAL_SOLAR_ECLIPSE, jd),
        'post_natal_lunar_eclipse': ephemeris.eclipse(chart.POST_NATAL_LUNAR_ECLIPSE, jd),
    }

    for key, eph_object in data.items():
        # Convert ecliptical longitude to sign-based
        object = eph_object.copy()
        object['lon'] = position.sign_longitude(object)

        # Format properties to string to match astro.com front-end output
        for property_key in ('lon', 'lat', 'speed', 'dec'):
            if property_key in object:
                object[property_key] = convert.dec_to_string(object[property_key])

        # For eclipse dates
        if 'date' in astro[key]:
            object['date'] = date.to_datetime(object['jd'], *coords).strftime('%d %B')

        for property, value in astro[key].items():
            assert object[property] == value


def test_armc_get_data(coords, jd, astro, armc):
    setup.add_filepath(os.path.dirname(__file__))

    obliquity = ephemeris.obliquity(jd)

    data = {
        'asc': ephemeris.armc_angle(chart.ASC, armc, coords[0], obliquity, chart.PLACIDUS),
        'house_2': ephemeris.armc_house(chart.HOUSE2, armc, coords[0], obliquity, chart.PLACIDUS),
        'pof': ephemeris.armc_point(chart.PARS_FORTUNA, jd, armc, coords[0], obliquity, pars_fortuna_formula=calc.DAY_NIGHT_FORMULA),
    }

    for key, eph_object in data.items():
        # Convert ecliptical longitude to sign-based
        object = eph_object.copy()
        object['lon'] = position.sign_longitude(object)

        # Format properties to string to match astro.com front-end output
        for property_key in ('lon', 'lat', 'speed', 'dec'):
            if property_key in object:
                object[property_key] = convert.dec_to_string(object[property_key])

        for property, value in astro[key].items():
            assert object[property] == value


def test_moon_phase(jd):
    # Courtesy of https://stardate.org/nightsky/moon
    assert ephemeris.moon_phase(jd) == calc.THIRD_QUARTER             # third quarter = waning crescent


def test_obliquity(jd):
    # Courtesy of http://neoprogrammics.com/obliquity_of_the_ecliptic/Obliquity_Of_The_Ecliptic_Calculator.php
    assert ephemeris.obliquity(jd) == approx(23.4376888901)
    assert ephemeris.obliquity(jd, True) == approx(23.4392911408)


def test_is_daytime(jd, coords):
    # Sun above ascendant in astro.com chart visual
    assert ephemeris.is_daytime(jd, *coords)


def test_armc_is_daytime(jd, coords, armc):
    # Sun above ascendant in astro.com chart visual
    assert ephemeris.armc_is_daytime(jd, armc, coords[0], ephemeris.obliquity(jd))


def test_deltat(jd):
    # Courtesy of astro.com "Additional Tables"
    assert round(ephemeris.deltat(jd, True), 1) == 63.8



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    A very basic serializer class to use with the stock json module.

"""

from json import JSONEncoder


class ToJSON(JSONEncoder):
    def default(self, obj) -> dict | str | None:
        if hasattr(obj, 'to_json'):
            return obj.to_json()

        if hasattr(obj, '__dict__'):
            return {k: v for k, v in obj.__dict__.items() if k[0] != '_'}

        if hasattr(obj, '__str__'):
            return str(obj)

        return None



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module provides simple dignity and mutual reception data.

    Assuming only planets will be passed, these functions will return the main
    essential dignity states of any planet. Mutual receptions are also
    calculated for each diginity.

"""

from immanuel.const import dignities
from immanuel.setup import settings
from immanuel.tools import position


def ruler(object: dict) -> bool:
    """ Returns whether the passed planet is the ruler of its sign. """
    return object['index'] == settings.rulerships[position.sign(object)]


def exalted(object: dict) -> bool:
    """ Returns whether the passed planet is exalted within its sign. """
    return object['index'] == dignities.EXALTATIONS[position.sign(object)]


def triplicity_ruler(object: dict, is_daytime: bool) -> bool:
    """ Returns whether the passed planet is any type of triplicity ruler. """
    triplicities = settings.triplicities[position.sign(object)]

    if settings.include_participatory_triplicities:
        return object['index'] in (triplicities['day' if is_daytime else 'night'], triplicities.get('participatory'))
    else:
        return object['index'] == triplicities['day' if is_daytime else 'night']


def term_ruler(object: dict) -> bool:
    """ Returns whether the passed planet is the term ruler
    within its sign. """
    terms = settings.terms[position.sign(object)]

    if object['index'] not in terms:
        return False

    return terms[object['index']][0] <= position.sign_longitude(object) < terms[object['index']][1]


def face_ruler(object: dict) -> bool:
    """ Returns whether the passed planet is the decan ruler
    within its sign. """
    return object['index'] == dignities.FACE_RULERS[position.sign(object)][position.decan(object)-1]


def mutual_reception_ruler(object: dict, objects: dict) -> bool:
    """ Returns whether the passed planet is in mutual reception
    by rulership. """
    if ruler(object):
        return False

    object_sign = position.sign(object)
    object_sign_rulership = settings.rulerships[object_sign]

    return position.sign(objects[object_sign_rulership]) in _planet_signs(object, settings.rulerships)


def mutual_reception_exalted(object: dict, objects: dict) -> bool:
    """ Returns whether the passed planet is in mutual reception
    by exaltation. """
    if exalted(object):
        return False

    object_sign = position.sign(object)
    object_sign_exaltation = dignities.EXALTATIONS[object_sign]

    if object_sign_exaltation is None:
        return False

    return position.sign(objects[object_sign_exaltation]) in _planet_signs(object, dignities.EXALTATIONS)


def mutual_reception_triplicity_ruler(object: dict, objects: dict, is_daytime: bool) -> bool:
    """ Returns whether the passed planet is in mutual reception
    by any type of triplicity rulership. """
    if triplicity_ruler(object, is_daytime):
        return False

    key = 'day' if is_daytime else 'night'
    object_sign = position.sign(object)
    object_triplicities = settings.triplicities[object_sign]
    day_night_ruler_triplicities = settings.triplicities[position.sign(objects[object_triplicities[key]])]

    if object['index'] == day_night_ruler_triplicities[key]:
        return True

    if not settings.include_participatory_triplicities or 'participatory' not in object_triplicities:
        return False

    participatory_ruler_triplicities = settings.triplicities[position.sign(objects[object_triplicities['participatory']])]

    return object['index'] == participatory_ruler_triplicities['participatory']


def mutual_reception_term_ruler(object: dict, objects: dict) -> bool:
    """ Returns whether the passed planet is in mutual reception
    by term rulership. """
    if term_ruler(object):
        return False

    object_terms = settings.terms[position.sign(object)]

    for index, boundaries in object_terms.items():
        if boundaries[0] <= position.sign_longitude(object) < boundaries[1]:
            break

    term_ruler_terms = settings.terms[position.sign(objects[index])]

    return object['index'] in term_ruler_terms and term_ruler_terms[object['index']][0] <= position.sign_longitude(objects[index]) < term_ruler_terms[object['index']][1]


def mutual_reception_face_ruler(object: dict, objects: dict) -> bool:
    """ Returns whether the passed planet is in mutual reception
    by face rulership. """
    if face_ruler(object):
        return False

    face_ruler_object = dignities.FACE_RULERS[position.sign(object)][position.decan(object)-1]

    return object['index'] == dignities.FACE_RULERS[position.sign(objects[face_ruler_object])][position.decan(objects[face_ruler_object])-1]


def in_rulership_element(object: dict) -> bool:
    """ Returns whether the passed planet is in a sign that shares an element
    with its domicile. """
    if object['index'] not in settings.rulerships.values():
        return False

    return position.element(_planet_signs(object, settings.rulerships)[0]*30-1) == position.element(object)


def detriment(object: dict) -> bool:
    """ Returns whether the passed planet is in detriment within its sign. """
    return position.opposite_sign(object) in _planet_signs(object, settings.rulerships)


def fall(object: dict) -> bool:
    """ Returns whether the passed planet is in fall within its sign. """
    return position.opposite_sign(object) in _planet_signs(object, dignities.EXALTATIONS)


def all(object: dict, objects: dict, is_daytime: bool) -> dict:
    essential_dignities = {
        dignities.RULER: ruler(object),
        dignities.EXALTED: exalted(object),
        dignities.TRIPLICITY_RULER: triplicity_ruler(object, is_daytime),
        dignities.TERM_RULER: term_ruler(object),
        dignities.FACE_RULER: face_ruler(object),
    }

    mutual_reception_dignities = {
        dignities.MUTUAL_RECEPTION_RULER: mutual_reception_ruler(object, objects),
        dignities.MUTUAL_RECEPTION_EXALTED: mutual_reception_exalted(object, objects),
        dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER: mutual_reception_triplicity_ruler(object, objects, is_daytime),
        dignities.MUTUAL_RECEPTION_TERM_RULER: mutual_reception_term_ruler(object, objects),
        dignities.MUTUAL_RECEPTION_FACE_RULER: mutual_reception_face_ruler(object, objects),
    }

    minor_dignities = {
        dignities.IN_RULERSHIP_ELEMENT: in_rulership_element(object),
    }

    debilities = {
        dignities.DETRIMENT: detriment(object),
        dignities.FALL: fall(object),
        dignities.PEREGRINE: not any(essential_dignities.values()),
    }

    if settings.include_mutual_receptions:
        debilities[dignities.PEREGRINE] = debilities[dignities.PEREGRINE] and not any(mutual_reception_dignities.values())

    if debilities[dignities.PEREGRINE]:
        debilities[dignities.PEREGRINE] = not minor_dignities[dignities.IN_RULERSHIP_ELEMENT]

    return essential_dignities | mutual_reception_dignities | minor_dignities | debilities


def score(dignity_state: dict) -> int:
    """ Calculates a planet's dignity score based on settings. """
    return sum([v for k, v in settings.dignity_scores.items() if k in dignity_state and dignity_state[k]])


def _planet_signs(object: dict, table: dict) -> tuple:
    """ Returns the sign(s) a planet belongs to in {sign: planet} dicts. """
    return tuple(k for k, v in table.items() if v == object['index'])



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module calculates past and future transits relative
    to a passed Julian day.

    previous() and next() take two planets, a Julian day, and an aspect.
    The functions will then find & return the last/next Julian day
    before/after the passed one where the requested aspect took place.
    This calculation can be expensive for slow planets, and is currently
    not flexible enough to use with non-planetary objects or points.

    The previous/next new and full moon functions are designed to
    fast-rewind and fast-forward to a close approximation of each aspect
    before handing off to _find()'s loop. Since the Sun and Moon have
    relatively stable daily motions and never retrograde, these are the only
    two bodies predictable enough to safely perform this with.

    The eclipse functions hand off directly to pyswisseph's own functions.

"""

import math

import swisseph as swe

from immanuel.const import calc, chart
from immanuel.tools import ephemeris


PREVIOUS = 0
NEXT = 1

_SWE = {
    swe.ECL_TOTAL: chart.TOTAL,
    swe.ECL_ANNULAR: chart.ANNULAR,
    swe.ECL_PARTIAL: chart.PARTIAL,
    swe.ECL_ANNULAR_TOTAL: chart.ANNULAR_TOTAL,
    swe.ECL_PENUMBRAL: chart.PENUMBRAL,
}


def previous(first: int, second: int, jd: float, aspect: float) -> float:
    """ Returns the Julian day of the requested transit previous
    to the passed Julian day. """
    return _find(first, second, jd, aspect, PREVIOUS)


def next(first: int, second: int, jd: float, aspect: float) -> float:
    """ Returns the Julian day of the requested transit after
    the passed Julian day. """
    return _find(first, second, jd, aspect, NEXT)


def previous_new_moon(jd: float) -> float:
    """ Fast rewind to approximate conjunction. """
    sun = ephemeris.planet(chart.SUN, jd)
    moon = ephemeris.planet(chart.MOON, jd)
    distance = swe.difdegn(moon['lon'], sun['lon'])
    jd -= math.floor(distance) / math.ceil(calc.MEAN_MOTIONS[chart.MOON])
    return previous(chart.SUN, chart.MOON, jd, calc.CONJUNCTION)


def previous_full_moon(jd: float) -> float:
    """ Fast rewind to approximate opposition. """
    sun = ephemeris.planet(chart.SUN, jd)
    moon = ephemeris.planet(chart.MOON, jd)
    distance = swe.difdegn(moon['lon'], sun['lon']+180)
    jd -= math.floor(distance) / math.ceil(calc.MEAN_MOTIONS[chart.MOON])
    return previous(chart.SUN, chart.MOON, jd, calc.OPPOSITION)


def next_new_moon(jd: float) -> float:
    """ Fast forward to approximate conjunction. """
    sun = ephemeris.planet(chart.SUN, jd)
    moon = ephemeris.planet(chart.MOON, jd)
    distance = swe.difdegn(sun['lon'], moon['lon'])
    jd += math.floor(distance) / math.ceil(calc.MEAN_MOTIONS[chart.MOON])
    return next(chart.SUN, chart.MOON, jd, calc.CONJUNCTION)


def next_full_moon(jd: float) -> float:
    """ Fast forward to approximate opposition. """
    sun = ephemeris.planet(chart.SUN, jd)
    moon = ephemeris.planet(chart.MOON, jd)
    distance = swe.difdegn(sun['lon']+180, moon['lon'])
    jd += math.floor(distance) / math.ceil(calc.MEAN_MOTIONS[chart.MOON])
    return next(chart.SUN, chart.MOON, jd, calc.OPPOSITION)


def previous_solar_eclipse(jd: float) -> tuple:
    """ Returns the eclipse type and Julian date of the moment of maximum
    eclipse for the most recent global solar eclipse that occurred before the
    passed Julian date. """
    res, tret = swe.sol_eclipse_when_glob(jd, swe.FLG_SWIEPH, swe.ECL_ALLTYPES_SOLAR, True)
    return _eclipse_type(res), tret[0]


def previous_lunar_eclipse(jd: float) -> float:
    """ Returns the eclipse type and Julian date of the moment of maximum
    eclipse for the most recent lunar eclipse that occurred before the
    passed Julian date. """
    res, tret = swe.lun_eclipse_when(jd, swe.FLG_SWIEPH, swe.ECL_ALLTYPES_LUNAR, True)
    return _eclipse_type(res), tret[0]


def next_solar_eclipse(jd: float) -> float:
    """ Returns the eclipse type and Julian date of the moment of maximum
    eclipse for the next global solar eclipse that occurred after the
    passed Julian date. """
    res, tret = swe.sol_eclipse_when_glob(jd, swe.FLG_SWIEPH, swe.ECL_ALLTYPES_SOLAR)
    return _eclipse_type(res), tret[0]


def next_lunar_eclipse(jd: float) -> float:
    """ Returns the eclipse type and Julian date of the moment of maximum
    eclipse for the next lunar eclipse that occurred after the
    passed Julian date. """
    res, tret = swe.lun_eclipse_when(jd, swe.FLG_SWIEPH, swe.ECL_ALLTYPES_LUNAR)
    return _eclipse_type(res), tret[0]


def _find(first: int, second: int, jd: float, aspect: float, direction: int) -> float:
    """ Returns the Julian date of the previous/next requested aspect.
    Accurate to within one arc-second. """
    multiplier = 1 if direction == NEXT else -1

    while True:
        first_object = ephemeris.get(first, jd)
        second_object = ephemeris.get(second, jd)
        distance = abs(swe.difdeg2n(first_object['lon'], second_object['lon']))
        diff = abs(aspect - distance)

        if diff <= calc.MAX_ERROR:
            return jd

        add = 1 * multiplier
        speed = abs(max(first_object['speed'], second_object['speed']) - min(first_object['speed'], second_object['speed']))

        if diff < speed:
            add *= diff / 180

        jd += add


def _eclipse_type(swe_index: int) -> int:
    """ Returns the internal index of an eclipse type based on swisseph's
    bit flags. This clears the ECL_CENTRAL / ECL_NONCENTRAL bits from the
    end and maintains the simple eclipse type flag. """
    return _SWE[(swe_index >> 2) << 2]



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Basic midpoint calculations for chart objects. Two options exist for
    house-based objects such as the Vertex or primary angles: either allow
    them to be midpoint-calculated along with the other objects, or recalculate
    them based on a composite ARMC.

"""

import swisseph as swe

from immanuel.tools import ephemeris


def all(objects1: dict, objects2: dict, obliquity: float = None) -> dict:
    """ Takes two dicts of chart objects typically returned by the ephemeris
    module and returns the averaged data for both sets as a third dict of
    composite objects. """
    objects = {}

    for index, object in objects1.items():
        objects[index] = composite(object, objects2[index], obliquity)

    return objects


def composite(object1: dict, object2: dict, obliquity: float = None) -> dict:
    """ Given two chart objects typically returned by the ephemeris module,
    this function will return a composite object. """
    object = object1 | {
        'lon': swe.deg_midp(object1['lon'], object2['lon']),
        'speed': (object1['speed'] + object2['speed']) / 2,
    }

    if 'lat' in object1 and 'lat' in object2:
        object['lat'] = 0.0

    if 'dist' in object1 and 'dist' in object2:
        object['dist'] = 0.0

    if 'size' in object1 and 'size' in object2:
        object['size'] = (object1['size'] + object2['size']) / 2

    if 'dec' in object1 and obliquity is not None:
        object['dec'] = swe.cotrans((object['lon'], 0, 1), -obliquity)[1]

    return object


def obliquity(jd1: float, jd2: float, mean: bool = False) -> float:
    """ Returns the mean obliquity of two dates. """
    return (ephemeris.obliquity(jd1, mean) + ephemeris.obliquity(jd2, mean)) / 2



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    User-facing names for chart data objects.

"""

from immanuel.const import calc, chart, dignities


CHART_TYPES = {
    chart.NATAL: 'Natal',
    chart.SOLAR_RETURN: 'Solar Return',
    chart.PROGRESSED: 'Progressed',
    chart.COMPOSITE: 'Composite',
    chart.TRANSITS: 'Transits'
}

OBJECTS = {
    chart.HOUSE: 'House',
    chart.ANGLE: 'Angle',
    chart.PLANET: 'Planet',
    chart.ASTEROID: 'Asteroid',
    chart.POINT: 'Point',
    chart.ECLIPSE: 'Eclipse',
    chart.FIXED_STAR: 'Fixed star',
}

SIGNS = {
    chart.ARIES: 'Aries',
    chart.TAURUS: 'Taurus',
    chart.GEMINI: 'Gemini',
    chart.CANCER: 'Cancer',
    chart.LEO: 'Leo',
    chart.VIRGO: 'Virgo',
    chart.LIBRA: 'Libra',
    chart.SCORPIO: 'Scorpio',
    chart.SAGITTARIUS: 'Sagittarius',
    chart.CAPRICORN: 'Capricorn',
    chart.AQUARIUS: 'Aquarius',
    chart.PISCES: 'Pisces',
}

DECANS = {
    chart.DECAN1: '1st Decan',
    chart.DECAN2: '2nd Decan',
    chart.DECAN3: '3rd Decan',
}

ELEMENTS = {
    chart.FIRE: 'Fire',
    chart.EARTH: 'Earth',
    chart.AIR: 'Air',
    chart.WATER: 'Water',
}

MODALITIES = {
    chart.CARDINAL: 'Cardinal',
    chart.FIXED: 'Fixed',
    chart.MUTABLE: 'Mutable',
}

HOUSE_SYSTEMS = {
    chart.ALCABITUS: 'Alcabitus',
    chart.AZIMUTHAL: 'Azimuthal',
    chart.CAMPANUS: 'Campanus',
    chart.EQUAL: 'Equal House',
    chart.KOCH: 'Koch',
    chart.MERIDIAN: 'Meridian',
    chart.MORINUS: 'Morinus',
    chart.PLACIDUS: 'Placidus',
    chart.POLICH_PAGE: 'Polich Page',
    chart.PORPHYRIUS: 'Porphyrius',
    chart.REGIOMONTANUS: 'Regiomontanus',
    chart.VEHLOW_EQUAL: 'Vehlow Equal House',
    chart.WHOLE_SIGN: 'Whole Sign',
}

HOUSES = {
    chart.HOUSE1: '1st House',
    chart.HOUSE2: '2nd House',
    chart.HOUSE3: '3rd House',
    chart.HOUSE4: '4th House',
    chart.HOUSE5: '5th House',
    chart.HOUSE6: '6th House',
    chart.HOUSE7: '7th House',
    chart.HOUSE8: '8th House',
    chart.HOUSE9: '9th House',
    chart.HOUSE10: '10th House',
    chart.HOUSE11: '11th House',
    chart.HOUSE12: '12th House',
}

ANGLES = {
    chart.ASC: 'Asc',
    chart.DESC: 'Desc',
    chart.MC: 'MC',
    chart.IC: 'IC',
    chart.ARMC: 'ARMC',
}

PLANETS = {
    chart.SUN: 'Sun',
    chart.MOON: 'Moon',
    chart.MERCURY: 'Mercury',
    chart.VENUS: 'Venus',
    chart.MARS: 'Mars',
    chart.JUPITER: 'Jupiter',
    chart.SATURN: 'Saturn',
    chart.URANUS: 'Uranus',
    chart.NEPTUNE: 'Neptune',
    chart.PLUTO: 'Pluto',
}

ASTEROIDS = {
    chart.CHIRON: 'Chiron',
    chart.PHOLUS: 'Pholus',
    chart.CERES: 'Ceres',
    chart.PALLAS: 'Pallas',
    chart.JUNO: 'Juno',
    chart.VESTA: 'Vesta',
}

POINTS = {
    chart.NORTH_NODE: 'North Node',
    chart.SOUTH_NODE: 'South Node',
    chart.TRUE_NORTH_NODE: 'True North Node',
    chart.TRUE_SOUTH_NODE: 'True South Node',
    chart.VERTEX: 'Vertex',
    chart.LILITH: 'Lilith',
    chart.TRUE_LILITH: 'True Lilith',
    chart.SYZYGY: 'Syzygy',
    chart.PARS_FORTUNA: 'Part of Fortune',
}

ECLIPSES = {
    chart.PRE_NATAL_LUNAR_ECLIPSE: 'Pre-natal Lunar Eclipse',
    chart.PRE_NATAL_SOLAR_ECLIPSE: 'Pre-natal Solar Eclipse',
    chart.POST_NATAL_LUNAR_ECLIPSE: 'Post-natal Lunar Eclipse',
    chart.POST_NATAL_SOLAR_ECLIPSE: 'Post-natal Solar Eclipse',
}

ECLIPSE_TYPES = {
    chart.TOTAL: 'Total',
    chart.ANNULAR: 'Annular',
    chart.PARTIAL: 'Partial',
    chart.ANNULAR_TOTAL: 'Annular Total',
    chart.PENUMBRAL: 'Penumbral',
}

ASPECTS = {
    calc.CONJUNCTION: 'Conjunction',
    calc.OPPOSITION: 'Opposition',
    calc.SQUARE: 'Square',
    calc.TRINE: 'Trine',
    calc.SEXTILE: 'Sextile',
    calc.SEPTILE: 'Septile',
    calc.SEMISQUARE: 'Semisquare',
    calc.SESQUISQUARE: 'Sesquisquare',
    calc.SEMISEXTILE: 'Semisextile',
    calc.QUINCUNX: 'Quincunx',
    calc.QUINTILE: 'Quintile',
    calc.BIQUINTILE: 'Biquintile',
}

ASPECT_MOVEMENTS = {
    calc.APPLICATIVE: 'Applicative',
    calc.EXACT: 'Exact',
    calc.SEPARATIVE: 'Separative',
}

ASPECT_CONDITIONS = {
    calc.ASSOCIATE: 'Associate',
    calc.DISSOCIATE: 'Dissociate',
}

DIGNITIES = {
    dignities.RULER: 'Ruler',
    dignities.EXALTED: 'Exalted',
    dignities.TRIPLICITY_RULER: 'Triplicity Ruler',
    dignities.TERM_RULER: 'Term Ruler',
    dignities.FACE_RULER: 'Face Ruler',
    dignities.MUTUAL_RECEPTION_RULER: 'Ruler by mutual reception',
    dignities.MUTUAL_RECEPTION_EXALTED: 'Exalted by mutual reception',
    dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER: 'Triplicity Ruler by mutual reception',
    dignities.MUTUAL_RECEPTION_TERM_RULER: 'Term Ruler by mutual reception',
    dignities.MUTUAL_RECEPTION_FACE_RULER: 'Face Ruler by mutual reception',
    dignities.IN_RULERSHIP_ELEMENT: 'In rulership element',
    dignities.DETRIMENT: 'Detriment',
    dignities.FALL: 'Fall',
    dignities.PEREGRINE: 'Peregrine',
}

MOON_PHASES = {
    calc.NEW_MOON: 'New',
    calc.WAXING_CRESCENT: 'Waxing Crescent',
    calc.FIRST_QUARTER: 'First Quarter',
    calc.WAXING_GIBBOUS: 'Waxing Gibbous',
    calc.FULL_MOON: 'Full',
    calc.DISSEMINATING: 'Disseminating',
    calc.THIRD_QUARTER: 'Third Quarter',
    calc.BALSAMIC: 'Balsamic',
}

OBJECT_MOVEMENTS = {
    calc.DIRECT: 'Direct',
    calc.STATIONARY: 'Stationary',
    calc.RETROGRADE: 'Retrograde',
}

CHART_SHAPES = {
    calc.BUNDLE: 'Bundle',
    calc.BUCKET: 'Bucket',
    calc.BOWL: 'Bowl',
    calc.LOCOMOTIVE: 'Locomotive',
    calc.SEESAW: 'Seesaw',
    calc.SPLAY: 'Splay',
    calc.SPLASH: 'Splash',
}

PROGRESSION_METHODS = {
    calc.NAIBOD: 'Naibod',
    calc.SOLAR_ARC: 'Solar Arc',
    calc.DAILY_HOUSES: 'Daily Houses',
}



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Defines indices for the main chart objects supported by pyswisseph,
    along with custom-calculated objects based on pyswisseph. These
    constants are mainly used by the ephemeris module.

"""


""" Determine object type by dividing index by this. Since asteroids are passed
by number, this must be higher than the highest-numbered asteroid. """
TYPE_MULTIPLIER = 1000000

""" Supported chart types. """
NATAL = 1
SOLAR_RETURN = 2
PROGRESSED = 3
COMPOSITE = 4
TRANSITS = 5

""" Signs. """
ARIES = 1
TAURUS = 2
GEMINI = 3
CANCER = 4
LEO = 5
VIRGO = 6
LIBRA = 7
SCORPIO = 8
SAGITTARIUS = 9
CAPRICORN = 10
AQUARIUS = 11
PISCES = 12

""" Decans. """
DECAN1 = 1
DECAN2 = 2
DECAN3 = 3

""" Elements. """
FIRE = 1
EARTH = 2
AIR = 3
WATER = 4

""" Modalities. """
CARDINAL = 1
FIXED = 2
MUTABLE = 3

""" House systems. """
ALCABITUS = 101
AZIMUTHAL = 102
CAMPANUS = 103
EQUAL = 104
KOCH = 105
MERIDIAN = 106
MORINUS = 107
PLACIDUS = 108
POLICH_PAGE = 109
PORPHYRIUS = 110
REGIOMONTANUS = 111
VEHLOW_EQUAL = 112
WHOLE_SIGN = 113

""" Houses. """
HOUSE = 2 * TYPE_MULTIPLIER
HOUSE1 = HOUSE + 1
HOUSE2 = HOUSE + 2
HOUSE3 = HOUSE + 3
HOUSE4 = HOUSE + 4
HOUSE5 = HOUSE + 5
HOUSE6 = HOUSE + 6
HOUSE7 = HOUSE + 7
HOUSE8 = HOUSE + 8
HOUSE9 = HOUSE + 9
HOUSE10 = HOUSE + 10
HOUSE11 = HOUSE + 11
HOUSE12 = HOUSE + 12

""" Major chart angles. """
ANGLE = 3 * TYPE_MULTIPLIER
ASC = ANGLE + 1
DESC = ANGLE + 2
MC = ANGLE + 3
IC = ANGLE + 4
ARMC = ANGLE + 5

""" Planets. """
PLANET = 4 * TYPE_MULTIPLIER
SUN = PLANET + 1
MOON = PLANET + 2
MERCURY = PLANET + 3
VENUS = PLANET + 4
MARS = PLANET + 5
JUPITER = PLANET + 6
SATURN = PLANET + 7
URANUS = PLANET + 8
NEPTUNE = PLANET + 9
PLUTO = PLANET + 10

""" Major asteroids. """
ASTEROID = 5 * TYPE_MULTIPLIER
CHIRON = ASTEROID + 1
PHOLUS = ASTEROID + 2
CERES = ASTEROID + 3
PALLAS = ASTEROID + 4
JUNO = ASTEROID + 5
VESTA = ASTEROID + 6

""" Main calculated points. """
POINT = 6 * TYPE_MULTIPLIER
NORTH_NODE = POINT + 1
SOUTH_NODE = POINT + 2
TRUE_NORTH_NODE = POINT + 3
TRUE_SOUTH_NODE = POINT + 4
VERTEX = POINT + 5
LILITH = POINT + 6
TRUE_LILITH = POINT + 7
SYZYGY = POINT + 8
PARS_FORTUNA = POINT + 9

""" Fixed stars. """
FIXED_STAR = 7 * TYPE_MULTIPLIER

""" Pre & post-natal eclipses. """
ECLIPSE = 8 * TYPE_MULTIPLIER
PRE_NATAL_SOLAR_ECLIPSE = ECLIPSE + 1
PRE_NATAL_LUNAR_ECLIPSE = ECLIPSE + 2
POST_NATAL_SOLAR_ECLIPSE = ECLIPSE + 3
POST_NATAL_LUNAR_ECLIPSE = ECLIPSE + 4

""" Eclipse types. """
TOTAL = 1
ANNULAR = 2
PARTIAL = 3
ANNULAR_TOTAL = 4
PENUMBRAL = 5



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Inspects element and quadrant weighting from a dict of chart objects
    provided by the ephemeris module.

"""

from immanuel.const import chart
from immanuel.tools import position


def elements(objects: dict) -> dict:
    """ Returns data on the amount of chart objects
    belonging to each element. """
    weightings = {
        chart.FIRE: [],
        chart.EARTH: [],
        chart.AIR: [],
        chart.WATER: [],
    }

    for object in objects.values():
        weightings[position.element(object)].append(object['index'])

    return weightings


def modalities(objects: dict) -> dict:
    """ Returns data on the amount of chart objects
    belonging to each modality. """
    weightings = {
        chart.CARDINAL: [],
        chart.FIXED: [],
        chart.MUTABLE: [],
    }

    for object in objects.values():
        weightings[position.modality(object)].append(object['index'])

    return weightings


def quadrants(objects: dict, houses: dict) -> dict:
    """ Returns data on the amount of chart objects
    belonging to each of the chart's house quadrants. """
    weightings = {
        1: [],
        2: [],
        3: [],
        4: [],
    }

    for object in objects.values():
        house = position.house(object, houses)
        quadrant = int((house['number']-1) / 3) + 1
        weightings[quadrant].append(object['index'])

    return weightings



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Test dates compared against astro.com output.

"""

from datetime import datetime
from zoneinfo import ZoneInfo

from pytest import approx, fixture

from immanuel.const import calc, chart
from immanuel.tools import convert, date, ephemeris, forecast, position


@fixture
def coords():
    # San Diego coords as used by astro.com
    return [convert.string_to_dec(v) for v in ('32n43', '117w09')]

@fixture
def jd(coords):
    return date.to_jd('2000-01-01 10:00', *coords)

@fixture
def pjd():
    # Input JD for all progression tests - must be in UT
    return date.to_jd('2025-06-21')

@fixture
def astro():
    return {
        calc.DAILY_HOUSES: {
            chart.HOUSE1: {
                'sign': chart.VIRGO,
                'lon': '27°46\'45"',
            },
            chart.HOUSE2: {
                'sign': chart.LIBRA,
                'lon': '25°02\'08"',
            },
            chart.HOUSE3: {
                'sign': chart.SCORPIO,
                'lon': '25°26\'36"',
            },
            chart.HOUSE4: {
                'sign': chart.SAGITTARIUS,
                'lon': '27°36\'35"',
            },
            chart.HOUSE5: {
                'sign': chart.CAPRICORN,
                'lon': '29°46\'45"',
            },
            chart.HOUSE6: {
                'sign': chart.PISCES,
                'lon': '00°16\'09"',
            },
            chart.HOUSE7: {
                'sign': chart.PISCES,
                'lon': '27°46\'45"',
            },
            chart.HOUSE8: {
                'sign': chart.ARIES,
                'lon': '25°02\'08"',
            },
            chart.HOUSE9: {
                'sign': chart.TAURUS,
                'lon': '25°26\'36"',
            },
            chart.HOUSE10: {
                'sign': chart.GEMINI,
                'lon': '27°36\'35"',
            },
            chart.HOUSE11: {
                'sign': chart.CANCER,
                'lon': '29°46\'45"',
            },
            chart.HOUSE12: {
                'sign': chart.VIRGO,
                'lon': '00°16\'09"',
            },
        },
        calc.NAIBOD: {
            chart.HOUSE1: {
                'sign': chart.ARIES,
                'lon': '13°00\'29"',
            },
            chart.HOUSE2: {
                'sign': chart.TAURUS,
                'lon': '18°52\'16"',
            },
            chart.HOUSE3: {
                'sign': chart.GEMINI,
                'lon': '15°06\'02"',
            },
            chart.HOUSE4: {
                'sign': chart.CANCER,
                'lon': '07°57\'07"',
            },
            chart.HOUSE5: {
                'sign': chart.LEO,
                'lon': '01°57\'46"',
            },
            chart.HOUSE6: {
                'sign': chart.VIRGO,
                'lon': '01°58\'56"',
            },
            chart.HOUSE7: {
                'sign': chart.LIBRA,
                'lon': '13°00\'29"',
            },
            chart.HOUSE8: {
                'sign': chart.SCORPIO,
                'lon': '18°52\'16"',
            },
            chart.HOUSE9: {
                'sign': chart.SAGITTARIUS,
                'lon': '15°06\'02"',
            },
            chart.HOUSE10: {
                'sign': chart.CAPRICORN,
                'lon': '07°57\'07"',
            },
            chart.HOUSE11: {
                'sign': chart.AQUARIUS,
                'lon': '01°57\'46"',
            },
            chart.HOUSE12: {
                'sign': chart.PISCES,
                'lon': '01°58\'56"',
            },
        },
        calc.SOLAR_ARC: {
            chart.HOUSE1: {
                'sign': chart.ARIES,
                'lon': '17°32\'46"',
            },
            chart.HOUSE2: {
                'sign': chart.TAURUS,
                'lon': '22°20\'26"',
            },
            chart.HOUSE3: {
                'sign': chart.GEMINI,
                'lon': '18°01\'02"',
            },
            chart.HOUSE4: {
                'sign': chart.CANCER,
                'lon': '10°46\'59"',
            },
            chart.HOUSE5: {
                'sign': chart.LEO,
                'lon': '05°07\'03"',
            },
            chart.HOUSE6: {
                'sign': chart.VIRGO,
                'lon': '05°53\'34"',
            },
            chart.HOUSE7: {
                'sign': chart.LIBRA,
                'lon': '17°32\'46"',
            },
            chart.HOUSE8: {
                'sign': chart.SCORPIO,
                'lon': '22°20\'26"',
            },
            chart.HOUSE9: {
                'sign': chart.SAGITTARIUS,
                'lon': '18°01\'02"',
            },
            chart.HOUSE10: {
                'sign': chart.CAPRICORN,
                'lon': '10°46\'59"',
            },
            chart.HOUSE11: {
                'sign': chart.AQUARIUS,
                'lon': '05°07\'03"',
            },
            chart.HOUSE12: {
                'sign': chart.PISCES,
                'lon': '05°53\'34"',
            },
        },
    }


def test_solar_return_lon(jd):
    """ We use approx() here since the longitudes match
    to ~8 decimal places, but the longitudes maintain 13 places. """
    sr_jd = forecast.solar_return(jd, 2030)
    natal_sun = ephemeris.planet(chart.SUN, jd)
    sr_sun = ephemeris.planet(chart.SUN, sr_jd)
    assert natal_sun['lon'] == approx(sr_sun['lon'])


def test_solar_return_date(jd):
    """ Solar return date copied from astro.com """
    sr_jd = forecast.solar_return(jd, 2030)
    assert round(sr_jd + ephemeris.deltat(sr_jd), 6) == 2462502.521823


def test_progression_date(jd, pjd, coords):
    """ Progressed date copied from astro.com which returns UT date.
    Since the progressed date is always the same whatever method we
    use to calculate the houses, we can use any available method
    for this test. """
    progressed_jd = forecast.progression(jd, *coords, pjd, chart.PLACIDUS, calc.NAIBOD)[forecast.JD]
    assert round(progressed_jd + ephemeris.deltat(progressed_jd), 6) == 2451570.719456


def test_progression(jd, pjd, coords, astro):
    """ Test all available types of house progression. """
    for method, results in astro.items():
        progressed_jd, progressed_armc_lon = forecast.progression(jd, *coords, pjd, chart.PLACIDUS, method)

        for index, data in results.items():
            house = ephemeris.armc_house(index, progressed_armc_lon, coords[0], ephemeris.obliquity(progressed_jd), chart.PLACIDUS)
            sign = position.sign(house)
            lon = position.sign_longitude(house)
            assert sign == data['sign']
            assert convert.dec_to_string(lon) == data['lon']



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Provides a simple set of default settings that can be overridden.
    Also allows filepath(s) to ephermeris files to be changed or added.

"""

import os
from typing import Any

import swisseph as swe

from immanuel.const import calc, chart, data, dignities


_file_path = None

_settings = {}


""" Data that should be included for each chart type's output. """
_settings['chart_data'] = {
    chart.NATAL: [
        data.NATIVE,
        data.HOUSE_SYSTEM,
        data.SHAPE,
        data.DIURNAL,
        data.MOON_PHASE,
        data.OBJECTS,
        data.HOUSES,
        data.ASPECTS,
        data.WEIGHTINGS,
    ],
    chart.SOLAR_RETURN: [
        data.NATIVE,
        data.SOLAR_RETURN_YEAR,
        data.SOLAR_RETURN_DATE_TIME,
        data.HOUSE_SYSTEM,
        data.SHAPE,
        data.DIURNAL,
        data.MOON_PHASE,
        data.OBJECTS,
        data.HOUSES,
        data.ASPECTS,
        data.WEIGHTINGS,
    ],
    chart.PROGRESSED: [
        data.NATIVE,
        data.PROGRESSION_DATE_TIME,
        data.PROGRESSED_DATE_TIME,
        data.PROGRESSION_METHOD,
        data.HOUSE_SYSTEM,
        data.SHAPE,
        data.DIURNAL,
        data.MOON_PHASE,
        data.OBJECTS,
        data.HOUSES,
        data.ASPECTS,
        data.WEIGHTINGS,
    ],
    chart.COMPOSITE: [
        data.NATIVE,
        data.PARTNER,
        data.HOUSE_SYSTEM,
        data.SHAPE,
        data.DIURNAL,
        data.MOON_PHASE,
        data.OBJECTS,
        data.HOUSES,
        data.ASPECTS,
        data.WEIGHTINGS,
    ],
    chart.TRANSITS: [
        data.NATIVE,
        data.HOUSE_SYSTEM,
        data.SHAPE,
        data.DIURNAL,
        data.MOON_PHASE,
        data.OBJECTS,
        data.HOUSES,
        data.ASPECTS,
        data.WEIGHTINGS,
    ],
}


""" Default coordinates when none are supplied. Currently points to
the GMT prime meridian. """
_settings['default_latitude'] = 51.4779

_settings['default_longitude'] = -0.0015


""" House system as supported by pyswisseph. """
_settings['house_system'] = chart.PLACIDUS


""" Which planets, points etc. to show. """
_settings['objects'] = [
    chart.ASC, chart.DESC, chart.MC, chart.IC,
    chart.TRUE_NORTH_NODE, chart.TRUE_SOUTH_NODE,
    chart.VERTEX, chart.PARS_FORTUNA,
    chart.TRUE_LILITH,
    chart.SUN, chart.MOON, chart.MERCURY, chart.VENUS, chart.MARS,
    chart.JUPITER, chart.SATURN, chart.URANUS, chart.NEPTUNE, chart.PLUTO,
    chart.CHIRON,
]


""" Which planets, points etc. to use in chart shape calculations. """
_settings['chart_shape_objects'] = [
    chart.SUN, chart.MOON, chart.MERCURY, chart.VENUS, chart.MARS,
    chart.JUPITER, chart.SATURN, chart.URANUS, chart.NEPTUNE, chart.PLUTO,
]


""" Which aspects to calculate. """
_settings['aspects'] = [
    calc.CONJUNCTION, calc.OPPOSITION, calc.SQUARE, calc.TRINE, calc.SEXTILE,
    calc.QUINCUNX,
]


""" Rules for what chart objects can initiate or receive which aspects. """
_settings['default_aspect_rule'] = {
    'initiate': _settings['aspects'],
    'receive': _settings['aspects'],
}

_settings['planet_aspect_rule'] = {
    'initiate': _settings['aspects'],
    'receive': _settings['aspects'],
}

_settings['point_aspect_rule'] = {
    'initiate': (calc.CONJUNCTION,),
    'receive': _settings['aspects'],
}

_settings['aspect_rules'] = {
    chart.ASC: _settings['point_aspect_rule'],
    chart.DESC: _settings['point_aspect_rule'],
    chart.MC: _settings['point_aspect_rule'],
    chart.IC: _settings['point_aspect_rule'],

    chart.SUN: _settings['planet_aspect_rule'],
    chart.MOON: _settings['planet_aspect_rule'],
    chart.MERCURY: _settings['planet_aspect_rule'],
    chart.VENUS: _settings['planet_aspect_rule'],
    chart.MARS: _settings['planet_aspect_rule'],
    chart.JUPITER: _settings['planet_aspect_rule'],
    chart.SATURN: _settings['planet_aspect_rule'],
    chart.URANUS: _settings['planet_aspect_rule'],
    chart.NEPTUNE: _settings['planet_aspect_rule'],
    chart.PLUTO: _settings['planet_aspect_rule'],

    chart.NORTH_NODE: _settings['point_aspect_rule'],
    chart.SOUTH_NODE: _settings['point_aspect_rule'],
    chart.TRUE_NORTH_NODE: _settings['point_aspect_rule'],
    chart.TRUE_SOUTH_NODE: _settings['point_aspect_rule'],
    chart.SYZYGY: _settings['point_aspect_rule'],
    chart.PARS_FORTUNA: _settings['point_aspect_rule'],
    chart.VERTEX: _settings['point_aspect_rule'],
    chart.LILITH: _settings['point_aspect_rule'],
    chart.TRUE_LILITH: _settings['point_aspect_rule'],
}


""" Orbs for chart objects and their aspects. """
_settings['default_orb'] = 1.0

_settings['exact_orb'] = 0.3

_settings['orb_calculation'] = calc.MEAN

_settings['planet_orbs'] = {
    calc.CONJUNCTION: 10.0,
    calc.OPPOSITION: 10.0,
    calc.SQUARE: 10.0,
    calc.TRINE: 10.0,
    calc.SEXTILE: 6.0,
    calc.SEPTILE: 3.0,
    calc.SEMISQUARE: 3.0,
    calc.SESQUISQUARE: 3.0,
    calc.SEMISEXTILE: 3.0,
    calc.QUINCUNX: 3.0,
    calc.QUINTILE: 2.0,
    calc.BIQUINTILE: 2.0,
}

_settings['point_orbs'] = {
    calc.CONJUNCTION: 0.0,
    calc.OPPOSITION: 0.0,
    calc.SQUARE: 0.0,
    calc.TRINE: 0.0,
    calc.SEXTILE: 0.0,
    calc.SEPTILE: 0.0,
    calc.SEMISQUARE: 0.0,
    calc.SESQUISQUARE: 0.0,
    calc.SEMISEXTILE: 0.0,
    calc.QUINCUNX: 0.0,
    calc.QUINTILE: 0.0,
    calc.BIQUINTILE: 0.0,
}

_settings['orbs'] = {
    chart.ASC: _settings['planet_orbs'],
    chart.DESC: _settings['planet_orbs'],
    chart.MC: _settings['planet_orbs'],
    chart.IC: _settings['planet_orbs'],

    chart.SUN: _settings['planet_orbs'],
    chart.MOON: _settings['planet_orbs'],
    chart.MERCURY: _settings['planet_orbs'],
    chart.VENUS: _settings['planet_orbs'],
    chart.MARS: _settings['planet_orbs'],
    chart.JUPITER: _settings['planet_orbs'],
    chart.SATURN: _settings['planet_orbs'],
    chart.URANUS: _settings['planet_orbs'],
    chart.NEPTUNE: _settings['planet_orbs'],
    chart.PLUTO: _settings['planet_orbs'],

    chart.NORTH_NODE: _settings['point_orbs'],
    chart.SOUTH_NODE: _settings['point_orbs'],
    chart.TRUE_NORTH_NODE: _settings['point_orbs'],
    chart.TRUE_SOUTH_NODE: _settings['point_orbs'],
    chart.SYZYGY: _settings['point_orbs'],
    chart.PARS_FORTUNA: _settings['point_orbs'],
    chart.VERTEX: _settings['point_orbs'],
    chart.LILITH: _settings['point_orbs'],
    chart.TRUE_LILITH: _settings['point_orbs'],
}


""" Orb for calculating chart shapes. """
_settings['chart_shape_orb'] = 5.0


""" MC progression formula for secondary progressions. """
_settings['mc_progression_method'] = calc.NAIBOD


""" Part of Fortune formula. """
_settings['pars_fortuna_formula'] = calc.DAY_NIGHT_FORMULA


""" Dignity settings. """
_settings['rulerships'] = dignities.MODERN_RULERSHIPS

_settings['triplicities'] = dignities.PTOLEMAIC_TRIPLICITIES

_settings['terms'] = dignities.EGYPTIAN_TERMS

_settings['include_participatory_triplicities'] = False

_settings['include_mutual_receptions'] = True

_settings['dignity_scores'] = {
    dignities.RULER: 5,
    dignities.EXALTED: 4,
    dignities.TRIPLICITY_RULER: 3,
    dignities.TERM_RULER: 2,
    dignities.FACE_RULER: 1,
    dignities.MUTUAL_RECEPTION_RULER: 5,
    dignities.MUTUAL_RECEPTION_EXALTED: 4,
    dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER: 3,
    dignities.MUTUAL_RECEPTION_TERM_RULER: 2,
    dignities.MUTUAL_RECEPTION_FACE_RULER: 1,
    dignities.DETRIMENT: -5,
    dignities.FALL: -4,
    dignities.PEREGRINE: -5,
}


def set_filepath(path) -> None:
    """ Set the main file path for ephemeris files. """
    global _file_path
    _file_path = path
    swe.set_ephe_path(_file_path)


def add_filepath(path) -> None:
    """ Append a file path to the main one for custom extras. """
    global _file_path
    extra_path = f'{os.pathsep}{path}'

    if _file_path.endswith(extra_path):
        return

    _file_path += extra_path
    swe.set_ephe_path(_file_path)


class settings(type):
    """ Very simple class to keep track of settings. Default settings are preset
    and can be overridden in bulk with set() or singly attribute-style. """
    def dict(self) -> dict:
        return _settings

    def set(self, settings: dict) -> None:
        _settings.update(settings)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in _settings:
            self.set({
                name: value
            })

    def __getattr__(self, name) -> Any:
        return _settings[name]


class settings(metaclass=settings):
    pass



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Courtesy of timeanddate.com.
    Moon phase data: https://www.timeanddate.com/moon/phases
    Eclipse data: https://www.timeanddate.com/eclipse/list.html

"""


from datetime import datetime

from pytest import fixture

from immanuel.const import calc, chart
from immanuel.tools import convert, date, find


@fixture
def coords():
    # San Diego coords as used by astro.com
    return [convert.string_to_dec(v) for v in ('32n43', '117w09')]

@fixture
def jd(coords):
    return date.to_jd('2000-01-01 10:00', *coords)


def test_previous(jd, coords):
    # Check for previous Sun / Moon conjunction so we can use the same
    # test date/time as test_previous_new_moon()
    tr_jd = find.previous(chart.SUN, chart.MOON, jd, calc.CONJUNCTION)
    tr_dt = date.to_datetime(tr_jd, *coords)
    assert tr_dt.strftime('%Y-%m-%d %H:%M') == '1999-12-07 14:31'


def test_next(jd, coords):
    # Check for next Sun / Moon conjunction so we can use the same
    # test date/time as test_next_new_moon()
    tr_jd = find.next(chart.SUN, chart.MOON, jd, calc.CONJUNCTION)
    tr_dt = date.to_datetime(tr_jd, *coords)
    assert tr_dt.strftime('%Y-%m-%d %H:%M') == '2000-01-06 10:13'


def test_previous_new_moon(jd, coords):
    # https://www.timeanddate.com/moon/phases/?year=1999
    nm_jd = find.previous_new_moon(jd)
    nm_dt = date.to_datetime(nm_jd, *coords)
    assert nm_dt.strftime('%Y-%m-%d %H:%M') == '1999-12-07 14:31'


def test_previous_full_moon(jd, coords):
    # https://www.timeanddate.com/moon/phases/?year=1999
    nm_jd = find.previous_full_moon(jd)
    nm_dt = date.to_datetime(nm_jd, *coords)
    assert nm_dt.strftime('%Y-%m-%d %H:%M') == '1999-12-22 09:31'


def test_next_new_moon(jd, coords):
    # https://www.timeanddate.com/moon/phases/?year=2000
    nm_jd = find.next_new_moon(jd)
    nm_dt = date.to_datetime(nm_jd, *coords)
    assert nm_dt.strftime('%Y-%m-%d %H:%M') == '2000-01-06 10:13'


def test_next_full_moon(jd, coords):
    # https://www.timeanddate.com/moon/phases/?year=2000
    nm_jd = find.next_full_moon(jd)
    nm_dt = date.to_datetime(nm_jd, *coords)
    assert nm_dt.strftime('%Y-%m-%d %H:%M') == '2000-01-20 20:40'


def test_previous_solar_eclipse(jd, coords):
    # https://www.timeanddate.com/eclipse/list.html?starty=1990
    # https://www.timeanddate.com/eclipse/solar/1999-august-11
    ec_type, ec_jd = find.previous_solar_eclipse(jd)
    ec_dt = date.to_datetime(ec_jd, *coords)
    assert ec_type == chart.TOTAL
    assert ec_dt.strftime('%Y-%m-%d %H:%M') == '1999-08-11 04:03'


def test_previous_lunar_eclipse(jd, coords):
    # https://www.timeanddate.com/eclipse/list.html?starty=1990
    # https://www.timeanddate.com/eclipse/lunar/1999-july-28
    ec_type, ec_jd = find.previous_lunar_eclipse(jd)
    ec_dt = date.to_datetime(ec_jd, *coords)
    assert ec_type == chart.PARTIAL
    assert ec_dt.strftime('%Y-%m-%d %H:%M') == '1999-07-28 04:33'


def test_next_solar_eclipse(jd, coords):
    # https://www.timeanddate.com/eclipse/list.html?starty=2000
    # https://www.timeanddate.com/eclipse/solar/2000-february-5
    ec_type, ec_jd = find.next_solar_eclipse(jd)
    ec_dt = date.to_datetime(ec_jd, *coords)
    assert ec_type == chart.PARTIAL
    assert ec_dt.strftime('%Y-%m-%d %H:%M') == '2000-02-05 04:49'


def test_next_lunar_eclipse(jd, coords):
    # https://www.timeanddate.com/eclipse/list.html?starty=2000
    # https://www.timeanddate.com/eclipse/lunar/2000-january-21
    ec_type, ec_jd = find.next_lunar_eclipse(jd)
    ec_dt = date.to_datetime(ec_jd, *coords)
    assert ec_type == chart.TOTAL
    assert ec_dt.strftime('%Y-%m-%d %H:%M') == '2000-01-20 20:43'



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Gregorian UT / Julian Day conversions ran against
    figures from https://ssd.jpl.nasa.gov/tools/jdc/#/cd

"""

from datetime import datetime
from zoneinfo import ZoneInfo

from pytest import approx, fixture

from immanuel.tools import date


@fixture
def gmt_date():
    return datetime(2000, 1, 1, 18, tzinfo=ZoneInfo('Europe/London'))

@fixture
def pst_date():
    return datetime(2000, 1, 1, 10, tzinfo=ZoneInfo('America/Los_Angeles'))

@fixture
def str_gmt_date():
    return '2000-01-01 18:00:00'

@fixture
def str_pst_date():
    return '2000-01-01 10:00:00'

@fixture
def ambiguous_date():
    return datetime(2022, 11, 6, 1, 30, tzinfo=ZoneInfo('America/Los_Angeles'))

@fixture
def gmt_coords():
    return 51.509865, -0.118092     # London lat/lon

@fixture
def pst_coords():
    return 32.715736, -117.161087   # San Diego lat/lon

@fixture
def jd():
    return 2451545.25               # 2000-01-01 18:00 UT


def test_timezone_gmt(gmt_coords):
    assert date.timezone(*gmt_coords) == 'Europe/London'


def test_timezone_pst(pst_coords):
    assert date.timezone(*pst_coords) == 'America/Los_Angeles'


def test_localize(pst_coords):
    dt = datetime(2000, 1, 1, 18)
    aware = date.localize(dt, *pst_coords)
    assert aware.tzinfo == ZoneInfo('America/Los_Angeles')


def test_localize_dst(ambiguous_date, pst_coords):
    dt_no_dst = date.localize(ambiguous_date, *pst_coords, False)
    dt_dst = date.localize(ambiguous_date, *pst_coords, True)
    jd_no_dst = date.to_jd(dt_no_dst)
    jd_dst = date.to_jd(dt_dst)
    assert dt_no_dst.hour == dt_dst.hour
    assert jd_no_dst - jd_dst == approx(1/24)


def test_ambiguous(ambiguous_date, pst_date):
    assert date.ambiguous(ambiguous_date) == True
    assert date.ambiguous(pst_date) == False


def test_to_jd(str_pst_date, gmt_date, pst_date, pst_coords, jd):
    jd_original = date.to_jd(jd)
    jd_from_str = date.to_jd(str_pst_date, *pst_coords)
    jd_gmt = date.to_jd(gmt_date)
    jd_pst = date.to_jd(pst_date)
    assert jd_original == jd
    assert jd_from_str == jd
    assert jd_gmt == jd
    assert jd_pst == jd


def test_to_datetime_gmt(gmt_date, str_gmt_date, gmt_coords, jd):
    dt_original = date.to_datetime(gmt_date)
    assert dt_original.year == 2000
    assert dt_original.month == 1
    assert dt_original.day == 1
    assert dt_original.hour == 18
    assert dt_original.minute == 0
    assert dt_original.second == 0
    assert dt_original.tzinfo == ZoneInfo('Europe/London')

    dt_from_str = date.to_datetime(str_gmt_date, *gmt_coords)
    assert dt_from_str.year == 2000
    assert dt_from_str.month == 1
    assert dt_from_str.day == 1
    assert dt_from_str.hour == 18
    assert dt_from_str.minute == 0
    assert dt_from_str.second == 0
    assert dt_from_str.tzinfo == ZoneInfo('Europe/London')

    dt_from_jd = date.to_datetime(jd, *gmt_coords)
    assert dt_from_jd.year == 2000
    assert dt_from_jd.month == 1
    assert dt_from_jd.day == 1
    assert dt_from_jd.hour == 18
    assert dt_from_jd.minute == 0
    assert dt_from_jd.second == 0
    assert dt_from_jd.tzinfo == ZoneInfo('Europe/London')

    utc_dt_from_jd = date.to_datetime(jd)
    assert utc_dt_from_jd.year == 2000
    assert utc_dt_from_jd.month == 1
    assert utc_dt_from_jd.day == 1
    assert utc_dt_from_jd.hour == 18
    assert utc_dt_from_jd.minute == 0
    assert utc_dt_from_jd.second == 0
    assert utc_dt_from_jd.tzinfo == ZoneInfo('UTC')


def test_to_datetime_pst(pst_date, str_pst_date, pst_coords, jd):
    dt_original = date.to_datetime(pst_date)
    assert dt_original.year == 2000
    assert dt_original.month == 1
    assert dt_original.day == 1
    assert dt_original.hour == 10
    assert dt_original.minute == 0
    assert dt_original.second == 0
    assert dt_original.tzinfo == ZoneInfo('America/Los_Angeles')

    dt_from_str = date.to_datetime(str_pst_date, *pst_coords)
    assert dt_from_str.year == 2000
    assert dt_from_str.month == 1
    assert dt_from_str.day == 1
    assert dt_from_str.hour == 10
    assert dt_from_str.minute == 0
    assert dt_from_str.second == 0
    assert dt_from_str.tzinfo == ZoneInfo('America/Los_Angeles')

    dt_from_jd = date.to_datetime(jd, *pst_coords)
    assert dt_from_jd.year == 2000
    assert dt_from_jd.month == 1
    assert dt_from_jd.day == 1
    assert dt_from_jd.hour == 10
    assert dt_from_jd.minute == 0
    assert dt_from_jd.second == 0
    assert dt_from_jd.tzinfo == ZoneInfo('America/Los_Angeles')



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module provides simple helper functions for converting
    between location-specific Gregorian dates and times, and universal
    Julian dates.

    The conversion functions essentially wrap pyswisseph's julday() and
    revjul() to work with date/times in UT, but they take into account
    timezones based on lat/lon coordinates for the purposes of time offsets.
    This means that datetime objects expressing UT times will be zoned as UTC.

"""

from datetime import datetime
from zoneinfo import ZoneInfo

import swisseph as swe
from dateutil import tz
from timezonefinder import TimezoneFinder

from immanuel.tools import convert


def ambiguous(dt: datetime) -> bool:
    """ Returns whether an aware datetime is ambiguous. """
    return tz.datetime_ambiguous(dt)


def timezone(lat: float, lon: float) -> str:
    """ Returns a timezone string based on decimal lat/lon coordinates. """
    return TimezoneFinder().timezone_at(lat=lat, lng=lon)


def localize(dt: datetime, lat: float, lon: float, is_dst: bool = None) -> datetime:
    """ Localizes a naive datetime based on decimal lat/lon coordinates. """
    return dt.replace(tzinfo=ZoneInfo(timezone(lat, lon)), fold=1 if is_dst == False else 0)


def to_datetime(dt: str | float | datetime, lat: float = None, lon: float = None, is_dst: bool = None) -> datetime:
    """ Convert an unknown into a datetime. Unknowns can be either an
    ISO-formatted string, a Julian Date, or already a datetime. """
    no_coords = lat is None or lon is None
    if isinstance(dt, str):
        date_time = datetime.fromisoformat(dt)
        return date_time.replace(tzinfo=ZoneInfo('UTC')) if no_coords else localize(date_time, lat, lon, is_dst)
    if isinstance(dt, float):
        ut = swe.revjul(dt)
        time = convert.dec_to_dms(ut[3])[1:]
        date_time = datetime(*ut[:3], *time, tzinfo=ZoneInfo('UTC'))
        return date_time if no_coords else date_time.astimezone(ZoneInfo(timezone(lat, lon)))
    if isinstance(dt, datetime):
        if no_coords:
            return dt.replace(tzinfo=ZoneInfo('UTC')) if dt.tzinfo is None else dt
        else:
            return localize(dt, lat, lon, is_dst)
    return None


def to_jd(dt: str | float | datetime, lat: float = None, lon: float = None, is_dst: bool = None) -> float:
    """ Convert an unknown into a Julian date. Unknowns can be either an
    ISO-formatted string, already a Julian Date, or a datetime. """
    if isinstance(dt, float):
        return dt
    if isinstance(dt, str):
        date_time = datetime.fromisoformat(dt)
    elif isinstance(dt, datetime):
        date_time = dt
    else:
        return None

    if lat is not None and lon is not None:
        date_time = localize(date_time, lat, lon, is_dst)
    elif date_time.tzinfo is None:
        date_time = date_time.replace(tzinfo=ZoneInfo('UTC'))

    date_time_utc = date_time.astimezone(ZoneInfo('UTC'))
    hour = convert.dms_to_dec(('+', date_time_utc.hour, date_time_utc.minute, date_time_utc.second))
    return swe.julday(*date_time_utc.timetuple()[0:3], hour)



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module calculates aspects between chart objects based on
    the settings provided by the setup module.

    The functions also rely on the object data returned by the
    ephemeris module.

"""

import swisseph as swe

from immanuel.const import calc
from immanuel.setup import settings
from immanuel.tools import position


def between(object1: dict, object2: dict) -> dict:
    """ Returns any aspect between the two passed objects. """
    active, passive = (object1, object2) if abs(object1['speed']) > abs(object2['speed']) else (object2, object1)

    for aspect in settings.aspects:
        # Check aspect rules
        active_aspect_rule = settings.aspect_rules[active['index']] if active['index'] in settings.aspect_rules else settings.default_aspect_rule
        passive_aspect_rule = settings.aspect_rules[passive['index']] if passive['index'] in settings.aspect_rules else settings.default_aspect_rule

        if aspect not in active_aspect_rule['initiate'] or aspect not in passive_aspect_rule['receive']:
            return None

        # Get orbs
        active_orb = settings.orbs[active['index']][aspect] if active['index'] in settings.orbs else settings.default_orb
        passive_orb = settings.orbs[passive['index']][aspect] if passive['index'] in settings.orbs else settings.default_orb
        orb = ((active_orb + passive_orb) / 2) if settings.orb_calculation == calc.MEAN else max(active_orb, passive_orb)

        # Look for an aspect
        distance = swe.difdeg2n(passive['lon'], active['lon'])

        if aspect-orb <= abs(distance) <= aspect+orb:
            # Work out aspect information
            aspect_orb = abs(distance) - aspect
            exact_lon = swe.degnorm(passive['lon'] + (aspect if distance < 0 else -aspect))
            associate = position.sign(exact_lon) == position.sign(active)
            exact = exact_lon-settings.exact_orb <= active['lon'] <= exact_lon+settings.exact_orb
            applicative = not exact and ((aspect_orb < 0 if distance < 0 else aspect_orb > 0) or active['speed'] < -calc.STATION_SPEED)

            return {
                'active': active['index'],
                'passive': passive['index'],
                'aspect': aspect,
                'orb': orb,
                'distance': distance,
                'difference': aspect_orb,
                'movement': calc.EXACT if exact else calc.APPLICATIVE if applicative else calc.SEPARATIVE,
                'condition': calc.ASSOCIATE if associate else calc.DISSOCIATE,
            }

    return None


def for_object(object: dict, objects: dict, exclude_same: bool = True) -> dict:
    """ Returns all chart objects aspecting the passed chart object. If two
    separate sets of objects are being compared (eg. synastry) then
    exclude_self can be set to False to find aspects between the same
    object in both charts. """
    aspects = {}

    for index, check_object in objects.items():
        if exclude_same and index == object['index']:
            continue

        aspect = between(object, check_object)

        if aspect is not None:
            aspects[check_object['index']] = aspect

    return aspects


def all(objects: dict, exclude_same: bool = True) -> dict:
    """ Returns all aspects between the passed chart objects. """
    aspects = {}

    for index, object in objects.items():
        object_aspects = for_object(object, objects, exclude_same)

        if object_aspects:
            aspects[index] = object_aspects

    return aspects


def by_type(objects: dict, exclude_same: bool = True) -> dict:
    """ Returns all aspects between the passed chart objects keyed by
    aspect type. """
    aspects = {}

    for object in objects.values():
        object_aspects = for_object(object, objects, exclude_same)

        if object_aspects:
            for object_aspect in object_aspects.values():
                if object_aspect['aspect'] not in aspects:
                    aspects[object_aspect['aspect']] = []

                if object_aspect not in aspects[object_aspect['aspect']]:
                    aspects[object_aspect['aspect']].append(object_aspect)

    return aspects


def synastry(objects1: dict, objects2: dict, exclude_same: bool = False) -> dict:
    """ Returns all aspects between the two sets of passed chart objects. """
    aspects = {}

    for index, object in objects1.items():
        object_aspects = for_object(object, objects2, exclude_same)

        if object_aspects:
            aspects[index] = object_aspects

    return aspects



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Defines indices for the chart classes to define what data should be
    included for output. These are string rather than integer indices due to
    the way the chart classes use these to name their functions.

"""


NATIVE = 'native'
PARTNER = 'partner'
SOLAR_RETURN_YEAR = 'solar_return_year'
SOLAR_RETURN_DATE_TIME = 'solar_return_date_time'
PROGRESSION_DATE_TIME = 'progression_date_time'
PROGRESSED_DATE_TIME = 'progressed_date_time'
PROGRESSION_METHOD = 'progression_method'
HOUSE_SYSTEM = 'house_system'
SHAPE = 'shape'
DIURNAL = 'diurnal'
MOON_PHASE = 'moon_phase'
OBJECTS = 'objects'
ASPECTS = 'aspects'
HOUSES = 'houses'
WEIGHTINGS = 'weightings'



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module provides basic positional data for a given object's longitude.
    Since longitudes can either be extracted from objects or calculated
    directly, these functions will accept both an object and a float.

"""

import json

import swisseph as swe

from immanuel.const import chart


_house = {}


def sign(object: dict | float) -> int:
    """ Returns the index of the zodiac sign the
    passed object belongs to. """
    return int((object['lon'] if isinstance(object, dict) else object) / 30) + 1


def sign_longitude(object: dict | float) -> tuple:
    """ Returns the index of the zodiac sign the passed object
    belongs to, and the sign-specific longitude inside it. """
    return (object['lon'] if isinstance(object, dict) else object) % 30


def opposite_sign(object: dict | float) -> int:
    """ Returns the index of the zodiac sign opposite
    where the passed object belongs to. """
    sign_number = sign(object)
    return sign_number + (6 if sign_number <= 6 else -6)


def decan(object: dict | float) -> int:
    """ Returns which decan the passed object is within its sign. """
    return int((object['lon'] if isinstance(object, dict) else object) % 30) // 10 + 1


def house(object: dict | float, houses: dict) -> int:
    """ Given a object and a dict of houses from the ephemeris module, this
    returns which house the object is in. Basic dict caching is used. """
    lon = object['lon'] if isinstance(object, dict) else object
    key = json.dumps([lon, houses])

    if key in _house:
        return _house[key]

    for house in houses.values():
        lon_diff = swe.difdeg2n(lon, house['lon'])
        next_cusp_diff = swe.difdeg2n(house['lon'] + house['size'], house['lon'])

        if 0 <= lon_diff < next_cusp_diff:
            _house[key] = house
            return house


def opposite_house(object: dict | float, houses: dict) -> int:
    """ Given a object and a dict of houses from the ephemeris
    module, this returns the house opposite where the object is. """
    house_number = house((object['lon'] if isinstance(object, dict) else object), houses)['number']
    index = chart.HOUSE + house_number + (6 if house_number <= 6 else -6)
    return houses[index]


def element(object: dict | float) -> int:
    """ Returns the element associated with the sign
    which the passed object belongs to. """
    return int((object['lon'] if isinstance(object, dict) else object) / 30) % 4 + 1


def modality(object: dict | float) -> int:
    """ Returns the modality associated with the sign
    which the passed object belongs to. """
    return int((object['lon'] if isinstance(object, dict) else object) / 30) % 3 + 1



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Defines certain astrological and astronomical constants used for chart
    calculations. YEAR_DAYS and the sun's MEAN_MOTION are set to the same
    constants used by astro.com's calculations.

"""


from immanuel.const import chart

""" Aspects. """
CONJUNCTION = 0.0
OPPOSITION = 180.0
SQUARE = 90.0
TRINE = 120.0
SEXTILE = 60.0
SEPTILE = 51.43
SEMISQUARE = 45.0
SESQUISQUARE = 135.0
SEMISEXTILE = 30.0
QUINCUNX = 150.0
QUINTILE = 72.0
BIQUINTILE = 144.0

""" Calculations. """
MAX_ERROR = 0.000001                        # For precise exact conjunctions
STATION_SPEED = 0.0003                      # ~1 arc-second of movement
YEAR_DAYS = 365.24219893                    # Average days in a solar year (see swisseph swehouse.c)
J2000 = 2451545                             # Julian year 2000

""" Explicitly specify all of our planets as a set. """
PLANETS = {
    chart.SUN,
    chart.MOON,
    chart.MERCURY,
    chart.VENUS,
    chart.MARS,
    chart.JUPITER,
    chart.SATURN,
    chart.URANUS,
    chart.NEPTUNE,
    chart.PLUTO,
}

""" Mean daily planetary motions. """
MEAN_MOTIONS = {
    chart.SUN: 0.98564733,                  # See https://www.astro.com/cgi/h.cgi?f=gch&h=gch246&lang=e
    chart.MOON: 13.176389,
    chart.MERCURY: 1.383333,
    chart.VENUS: 1.2,
    chart.MARS: 0.524167,
    chart.JUPITER: 0.083056,
    chart.SATURN: 0.033611,
    chart.URANUS: 0.011667,
    chart.NEPTUNE: 0.006667,
    chart.PLUTO: 0.004167,
}

""" Moon phases. """
NEW_MOON = 45
WAXING_CRESCENT = 90
FIRST_QUARTER = 135
WAXING_GIBBOUS = 180
FULL_MOON = 225
DISSEMINATING = 270
THIRD_QUARTER = 315
BALSAMIC = 360

""" Object movements. """
DIRECT = 0
RETROGRADE = 1
STATIONARY = 2

""" MC progression formulae. """
NAIBOD = 0
SOLAR_ARC = 1
DAILY_HOUSES = 2

""" Part of Fortune formulae. """
DAY_FORMULA = 0
NIGHT_FORMULA = 1
DAY_NIGHT_FORMULA = 2

""" Composite Part of Fortune calculations. """
MIDPOINT = 0
COMPOSITE = 1

""" Orb calculation when two chart objects have different orbs. """
MEAN = 0
MAX = 1

""" Aspect information. """
APPLICATIVE = 0
EXACT = 1
SEPARATIVE = 2

ASSOCIATE = 0
DISSOCIATE = 1

""" Chart shapes. """
BUNDLE = 0
BUCKET = 1
BOWL = 2
LOCOMOTIVE = 3
SEESAW = 4
SPLAY = 5
SPLASH = 6
